"""
Copyright (c) 2015 Alister Chowdhury

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

http://alisterchowdhury.co.uk
"""
	
import maya.cmds as cmds
import pymel.core as pm
import maya.mel as mel
import maya.OpenMayaUI as omui
try:
	import cPickle as _pickle
except:
	import pickle as _pickle



class arc_rigbuilder(pm.uitypes.Window):
	
	wnd = ""
	
	
	@staticmethod
	def setupMenu():
		try:
			mel.eval("global string $gMainWindow;setParent $gMainWindow;menu -label \"Alister\" AlisterMenu;")
		except:
			pass
		try:
			mel.eval("setParent -menu AlisterMenu;menuItem -label \"Rig Builder\" -command \"python \\\"arc_rigbuilder.rigbuilder()\\\"\";")
		except:
			pass

	
	@staticmethod
	def rigbuilder():
		if not cmds.window(arc_rigbuilder.wnd, q=True, ex=True):
			arc_rigbuilder.wnd = arc_rigbuilder()
	
	
	def __init__(self):
		
		cmds.selectPref(trackSelectionOrder=True)
		self.joints = []
		self.bendy_iks = []
		self.pv_iks =[]
		self.geo_aimc = []
		self.geo_pc = []
		self.geo_sspc = []
		self.skipUpdate = False
		self.ignoreDelete = False
		self.ignoreIkUpdates = False
		self.ignoreGeoDUpdates = False
		self.cnjl_wnd = ""
		self.editik_wnd=""
		self.editdef_wnd=""
		
		
		
		self.setTitle("ARC Basic Rig Builder")
		
		
		cmds.menuBarLayout()
		
		cmds.menu("File")
		cmds.menuItem(label="New", command=self.newRig)
		cmds.menuItem(d=True)
		cmds.menuItem(label="Save Rig Data to File", command=self.saveRigData)
		cmds.menuItem(label="Load Rig Data from File", command=self.loadRigData)
		cmds.menuItem(d=True)
		cmds.menuItem(label="Preview Rig", command=self.previewRig)
		cmds.menuItem(label="Stop Preview Rig", command=self.deletePreviewRig)
		cmds.menuItem(d=True)
		cmds.menuItem(label="Bake Rig", command=self.bakeRig)
		cmds.menuItem(d=True)
		cmds.menuItem(label="Quit", command=self.close)
		
		
		cmds.menu("Joints")
		cmds.menuItem(label="Create Joint (Manual)", command=self.createManualJoint)
		cmds.menuItem(label="Create Joint (Inside of points)", command=self.cNewJointLoc)
		cmds.menuItem(label="Create # Joints (Spine tool)", command=self.createMultiplePoints)
		cmds.menuItem(label="Create # Joints (Copy to points)", command=self.createJointsToPoints)
		cmds.menuItem(d=True)
		cmds.menuItem(label="Mirror Joints", command=self.mirrorJoints)
		
		
		cmds.menu("IK Handle")
		cmds.menuItem(label="Create new Normal IK (RP + Pole Vector)", command=self.createNewPVIK)
		cmds.menuItem(label="Create new Bendy IK (Spline)", command=self.createNewBendyIK)
		cmds.menuItem(d=True)
		cmds.menuItem(label="Edit IKs", command=self.editIKs)
		
		
		cmds.menu("Geo Deformers")
		cmds.menuItem(label="Create new Eye Aimer (Aim Constraint)", command=self.createNewAimConstraint)
		cmds.menuItem(label="Create Point Controller (Parent Vertex)", command=self.createNewPointController)
		cmds.menuItem(label="Create new Softselect Controller (Soft Mod)", command=self.createNewSSPointController)
		cmds.menuItem(d=True)
		cmds.menuItem(label="Edit Geo Deformers", command=self.editDeformers)
		
		form = pm.formLayout()
		
		
		# The hassle I got using pm.scriptsTable ...
		self.jointstable = cmds.scriptTable(rows=0,columns=5, l=[(1,"name"), (2,"parent"), (3,"x"),(4,"y"),(5,"z")], cw=[(1,120),(2,120), (3,70), (4,70), (5,70)], ccc=self.updateJointsCCC)
		
		
		
		attachForm = [
		(self.jointstable, "left", 10),(self.jointstable, "right", 10),(self.jointstable,"top",10),(self.jointstable,"bottom",10)
		]
		
		pm.formLayout(form, edit=True, attachForm=attachForm)
		
		
		self.setWidth(475)
		
		# Callback when closed
		omui.MUiMessage.addUiDeletedCallback(self, self.uiDeleteCallback)
		self.show()
		
	
	def close(self, *args):
		try:
			cmds.deleteUI(arc_rigbuilder.wnd)
		except:
			pass
	
	
	def newRig(self, *args):
		self.cleanupEverything()
		self.joints = []
		self.bendy_iks = []
		self.pv_iks =[]
		self.geo_aimc = []
		self.geo_pc = []
		self.geo_sspc = []
		self.updateJointsList()
		self.updateIKsList()
	
	
	def cleanupScriptJobs(self):
		conds = cmds.scriptJob( listJobs=True )
		for cond in conds:
			# Check for any job with our prefix in it
			if "ARCBUILDER__" in cond:
				for i in range(len(cond)):
					
					# Extract the id
					if cond[i] == ":":
						break
				
				# Kill the job
				try:
					cmds.scriptJob(kill=int(cond[:i]))
				except:
					pass
	
	def cleanupEverything(self):
		igDel = self.ignoreDelete
		self.ignoreDelete = True
		self.deletePreviewRig()
		for v in self.joints:
			try:
				cmds.delete(v[2])
			except:
				pass
		try:
			cmds.delete("ARCBUILDER__*")
		except:
			pass
			
		self.cleanupScriptJobs()
		self.ignoreDelete = igDel
		
	
	# Cleanup all the stuff we've made
	def uiDeleteCallback(self, *args ):
		self.cleanupEverything()
		try:
			cmds.deleteUI(self.editik_wnd)
		except:
			pass
	
	def updateJointsCCC(self, r,c,v):
		
		#print [r, c, v, self.skipUpdate]
		if self.skipUpdate == True:
			return 1
			
		if(c < 3):
			if self.joints[r-1][c-1] != v:
				# Name changing
				if c == 1:
					try:
						for i in range(len(self.joints)):
							if(self.joints[i][0] == v):
								print "Error: Theres already something named that!"
								return 0
						
						oldname = self.joints[r-1][0]
						self.joints[r-1][2] = cmds.rename(self.joints[r-1][2], "ARCBUILDER__%s"%v)
						self.joints[r-1][0] = v
						
						# Change any joints which used this as a parent to the new parent name
						for i in range(len(self.joints)):
							if(self.joints[i][1] == oldname):
								self.joints[i][1] = v
						
						# Change any IK which used this aswell
						# This could still break the IK, if the IK is attached to a root
						for i in range(len(self.bendy_iks)):
							if(self.bendy_iks[i][1] == oldname):
								self.bendy_iks[i][1] = v
							
							if(self.bendy_iks[i][2] == oldname):
								self.bendy_iks[i][2] = v
						
						# Call for updates
						self.updateJointsList()
						return 1
					except:
						print "Error: Couldn't rename joint"
						return 0
				# Parent changing
				
				if c == 2:
					# wants to become root
					if v == "-":
						return 1
					
					for i in range(len(self.joints)):
						
						# Search to see if parent exists
						if(self.joints[i][1] == v):
						
							# No self parenting allowed
							if i == (r-1):
								print "Error: You can't parent something to itself, try changing it to '-', if you want to make it root"
								return 0
							
							# Parent found!
							self.joints[r-1][1] = v
							return 1
					
					print "Error: Parent not found!"
					return 0
			
			# Is just setting a name for the first time
			return 1
		try:
			fv = float(v)
		except:
			return 0
		
		
		xyz = cmds.xform(self.joints[r-1][2], q=True, ws=True, a=True, t=True)
		xyz[c-3] = fv
		cmds.xform(self.joints[r-1][2], ws=True, a=True, t=xyz)
		return 1
		
	
	def updateJointsList(self):
		l = len(self.joints)
		
		# Doing this stops scriptTable freaking out whenever someone changes a value and then adding a new joint
		# It seems to be because it tries to force a value on whatevers selectd, doing this removes everything, stopping that
		cmds.scriptTable(self.jointstable, e=True, r=0)
		cmds.scriptTable(self.jointstable, e=True, r=l)
		
		for i in range(l):
			cmds.scriptTable(self.jointstable, e=True, ci=[i+1,1], cv=str(self.joints[i][0])) 
			cmds.scriptTable(self.jointstable, e=True, ci=[i+1,2], cv=self.joints[i][1])
			xyz = cmds.xform(self.joints[i][2], q=True, ws=True, a=True, t=True)
			cmds.scriptTable(self.jointstable, e=True, ci=[i+1,3], cv=xyz[0])
			cmds.scriptTable(self.jointstable, e=True, ci=[i+1,4], cv=xyz[1])
			cmds.scriptTable(self.jointstable, e=True, ci=[i+1,5], cv=xyz[2])
	
	
	def cNewJointLoc(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
		
		
		self.cNewJointLoc_prev_mode = arc_rig_utils.getSelectMode()
		# Choose the vertexs that the joints will be in the middle of
		arc_rig_utils.setSelectMode(arc_rig_utils.SELECTMODE_COM)
		
		self.cnjl_wnd = cmds.window(title="Create New Joint Location", w=500, s=False)
		
		form = pm.formLayout()
		
		
		ok_btn = cmds.button("ok", label="OK",w=50, al="center", command=self._cNewJointLoc_callback)
		
		self.cNewJointLoc_name = cmds.textFieldGrp(l="Name:  	")
		self.cNewJointLoc_parent = cmds.optionMenu(l="Parent:	")
		cmds.menuItem(l="-")
		for e in self.joints:
			cmds.menuItem(l=e[0])
			
		try:
			cmds.optionMenu(self.cNewJointLoc_parent, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		
		lout = cmds.rowColumnLayout(numberOfColumns=1)
		cmds.text(width=500,label="Select the points you want the joint to be in the middle of" ,font="boldLabelFont",ww=True)
		cmds.text(width=500,label="Then select the name and parent joint (unless it's the root)" ,font="boldLabelFont",ww=True)
		cmds.text(width=500,label="Then press OK",font="boldLabelFont",ww=True)
		
		
		
		attachForm = [
		(lout, "left", 10),(lout, "right", 10),(lout,"top",10),
		(ok_btn, "bottom", 10)
		]
		
		attachControl =[
		(self.cNewJointLoc_name, "top", 20, lout),
		(self.cNewJointLoc_parent, "top", 10, self.cNewJointLoc_name),
		(ok_btn, "top", 10, self.cNewJointLoc_parent),
		]
		
		attachPosition =[
		(self.cNewJointLoc_name, "left", 0, 0),
		(self.cNewJointLoc_name, "right", 0, 50),
		(self.cNewJointLoc_parent, "left", 0, 12),
		(self.cNewJointLoc_parent, "right", 0, 50),
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		pm.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.cnjl_wnd)	
	
	def _cNewJointLoc_callback(self,a):
		nPos = arc_rig_utils.averagePosFromSelected()
		if nPos == None:
			print "Error: No points selected!"
			return
		
		name = cmds.textFieldGrp(self.cNewJointLoc_name, q=True, tx=True)
		parent = cmds.optionMenu(self.cNewJointLoc_parent, q=True, v=True)
		
		foundP = False
		for i in range(len(self.joints)):
			if self.joints[i][0] == name:
				print "Error: Name already exits!"
				return
			if self.joints[i][0] == parent:
				foundP = True
		
		if not foundP:
			if parent != "-":
				print "Error: Parent can't be found!"
				return
		
		self.addLocator(name, parent, nPos)
		self.updateJointsList()
		arc_rig_utils.setSelectMode(self.cNewJointLoc_prev_mode)
		cmds.deleteUI(self.cnjl_wnd)
	
	
	def addLocator(self, name, parent, pos):
		lc = cmds.spaceLocator(n="ARCBUILDER__%s"%name, p=(0,0,0))[0]
		arc_rig_utils.setColor(lc, 17)
		cmds.xform(lc, ws=True, a=True, t=pos)
		
		######################################################################
		# THIS NEEDS TO CHANGE IF I GIVE PEOPLE THE ABILITY TO DELETE JOINTS!#
		######################################################################
		cmds.scriptJob(attributeChange=["%s.tx"%lc, "arc_rigbuilder.wnd.locatorMoved(0,%i)"%len(self.joints)])
		cmds.scriptJob(attributeChange=["%s.ty"%lc, "arc_rigbuilder.wnd.locatorMoved(1,%i)"%len(self.joints)])
		cmds.scriptJob(attributeChange=["%s.tz"%lc, "arc_rigbuilder.wnd.locatorMoved(2,%i)"%len(self.joints)])
		cmds.scriptJob(attributeDeleted=["%s.tx"%lc, "arc_rigbuilder.wnd.locatorDeleted(%i)"%len(self.joints)], runOnce=True)
		
		self.joints.append([name, parent, lc])
		return lc
	
	def locatorMoved(self, ax, id, *args):
		#print [ax,id]
		
		xyz = cmds.xform(self.joints[id][2], q=True, ws=True, a=True, t=True)
		self.skipUpdate = True
		
		# Getting the currently selected cells, changing the selected cells to none, then changing them back
		# If this is not done, it will freak out and pass everything to the selected rows
		
		selcell = cmds.scriptTable(self.jointstable, q=True, selectedCells=True)
		cmds.scriptTable(self.jointstable, e=True, selectedRows=[])
		cmds.scriptTable(self.jointstable, e=True, ci=[id+1,3+ax], cv=xyz[ax])
		try:
			cmds.scriptTable(self.jointstable, e=True, selectedCells=selcell)
		except:
			# If nothing is selected it likes to throw out an exception
			pass
		self.skipUpdate = False
	
	
	def locatorDeleted(self, id, *args):
		
		if self.ignoreDelete:
			return
		
		self.ignoreDelete = True
		# Change everyone who has this locator as a parent and change their parent, to the locators parent
		name = self.joints[id][0]
		parent = self.joints[id][1]
		positions = []
		
		
		for jnt in self.joints:
			if jnt[1] == name:
				jnt[1] = parent
			
			pos = []
			try:
				pos = cmds.xform(jnt[2], q=True, ws=True, a=True, t=True)
			except:
				pass
				
			positions.append(pos)
		
		# Change any IK which used this aswell
		# This could still break the IK, if the IK is attached to a root
		for i in range(len(self.bendy_iks)):
			if(self.bendy_iks[i][1] == name):
				self.bendy_iks[i][1] = parent
			
			if(self.bendy_iks[i][2] == name):
				self.bendy_iks[i][2] = parent
		
		
		self.joints.pop(id)
		positions.pop(id)
		
		backup_pnts = self.joints[:]
		
		self.cleanupEverything()
		
		self.joints = []
		
		for i in range(len(backup_pnts)):
			
			name = backup_pnts[i][0]
			parent = backup_pnts[i][1]
			nPos = positions[i]
			self.addLocator(name,parent,nPos)
		
		self.updateJointsList()
		
		self.ignoreDelete = False
		
	
	
	def createMultiplePoints(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
					
		self.cCreateMultiple_prev_mode = arc_rig_utils.getSelectMode()
		# Choose the vertexs that the joints will be in the middle of
		arc_rig_utils.setSelectMode(arc_rig_utils.SELECTMODE_COM)
		
		self.cnjl_wnd = cmds.window(title="Create Spine", w=500, s=False)
		
		form = cmds.formLayout()
		
		
		ok_btn1 = cmds.button("ok1", label="OK", command=self._createMultiplePoints_callback, w=50)
		ok_btn2 = cmds.button("ok2", label="OK", command=self._createMultiplePoints_manual_callback, w=50)
		
		points_n_lbl1 = cmds.text(label="# of points")
		points_n_lbl2 = cmds.text(label="# of points")
		self.cmp_points_1 = cmds.intField("cmp_pnts_1", min=2, v=5)
		self.cmp_points_2 = cmds.intField("cmp_pnts_2", min=2, v=5)
		
		self.cmp_name_1 = cmds.textFieldGrp(l="Name: ", tx="spine")
		self.cmp_name_2 = cmds.textFieldGrp(l="Name: ", tx="spine")
		
		
		lout2 = cmds.text(width=500,label="Alternatively, enter them manually:" ,font="boldLabelFont",ww=True)
		
		sep1 = cmds.separator(h=10, style="out", w=500)
		
		pos1_ff = cmds.text(label="position 1:" , w=70)
		self.cmp_pos1X = cmds.floatField(pre=4, w=45)
		self.cmp_pos1Y = cmds.floatField(pre=4, w=45)
		self.cmp_pos1Z = cmds.floatField(pre=4, w=45)
		
		pos2_ff = cmds.text(label="position 2:", w=70)
		self.cmp_pos2X = cmds.floatField(pre=4, w=45)
		self.cmp_pos2Y = cmds.floatField(pre=4, w=45)
		self.cmp_pos2Z = cmds.floatField(pre=4, w=45)
		
		self.cmp_parent1 = cmds.optionMenu(l="Parent:	")
		cmds.menuItem(l="-")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.cmp_parent1, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
			
		self.cmp_parent2 = cmds.optionMenu(l="Parent:	")
		cmds.menuItem(l="-")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.cmp_parent2, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		
		lout = cmds.rowColumnLayout(numberOfColumns=1)
		
		cmds.text(width=500,label="First select the starting point, then the ending point" ,font="boldLabelFont",ww=True)
		cmds.text(width=500,label="Then press OK",font="boldLabelFont",ww=True)
		
				
		
		
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(lout2, "left", 0),(lout2, "right", 0),
		(ok_btn2, "bottom", 23),
		]
		
		attachControl =[
		(self.cmp_name_1, "top", 10, lout),
		(self.cmp_parent1, "top", 10, self.cmp_name_1),
		(points_n_lbl1, "top", 10, self.cmp_parent1),
		(self.cmp_points_1, "top", 7, self.cmp_parent1),(self.cmp_points_1, "left", 3, points_n_lbl1),
		(ok_btn1, "top", 15, self.cmp_points_1),
		(sep1, "top", 20, ok_btn1),
		(lout2, "top", 20, sep1),
		(pos1_ff, "top", 15, lout2),
		(self.cmp_pos1X, "top", 12, lout2),(self.cmp_pos1X, "left", 1, pos1_ff),
		(self.cmp_pos1Y, "top", 12, lout2),(self.cmp_pos1Y, "left", 1, self.cmp_pos1X),
		(self.cmp_pos1Z, "top", 12, lout2),(self.cmp_pos1Z, "left", 1, self.cmp_pos1Y),
		(pos2_ff, "top", 10, self.cmp_pos1Z),
		(self.cmp_pos2X, "top", 7, self.cmp_pos1Z),(self.cmp_pos2X, "left", 1, pos2_ff),
		(self.cmp_pos2Y, "top", 7, self.cmp_pos1Z),(self.cmp_pos2Y, "left", 1, self.cmp_pos2X),
		(self.cmp_pos2Z, "top", 7, self.cmp_pos1Z),(self.cmp_pos2Z, "left", 1, self.cmp_pos2Y),
		(self.cmp_name_2, "top", 10, self.cmp_pos2Z),
		(self.cmp_parent2, "top", 10, self.cmp_name_2),
		(points_n_lbl2, "top", 10, self.cmp_parent2),
		(self.cmp_points_2, "top", 7, self.cmp_parent2),(self.cmp_points_2, "left", 3, points_n_lbl2),
		(ok_btn2, "top", 15, self.cmp_points_2),
		]
		
		attachPosition =[
		(self.cmp_parent1, "left", 0, 13),
		(self.cmp_parent1, "right", 0, 50),
		(points_n_lbl1, "left", 0, 41),
		(ok_btn1, "left", 0, 45),
		(ok_btn1, "right", 0, 50),
		(self.cmp_parent2, "left", 0, 13),
		(self.cmp_parent2, "right", 0, 50),
		(points_n_lbl2, "left", 0, 41),
		(pos1_ff, "left", 0, 28),
		(pos2_ff, "left", 0, 28),
		(ok_btn2, "left", 0, 45),
		(ok_btn2, "right", 0, 50),
		]
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		
		cmds.showWindow(self.cnjl_wnd)	
	
	
	def _createMultiplePoints_callback(self,a):
		
		pnts = arc_rig_utils.getSelectedLocations()
		
		if(len(pnts) != 2):
			print "Error: Select exactly 2 points!"
			return
		
		nPnts = cmds.intField(self.cmp_points_1, q=True, v=True)
		prefix = cmds.textFieldGrp(self.cmp_name_2, q=True, tx=True)
		parent = cmds.optionMenu(self.cmp_parent2, q=True, v=True)
		p1 = pnts[0]
		p2 = pnts[1]
		
		self.addLocator("%s_1"%prefix, parent, p1)
		last_loc = "%s_1"%prefix
		
		for i in xrange(1,nPnts-1):
			pos = [arc_rig_utils.lerp(p1[0],p2[0], (float(i)/float(nPnts-1))), arc_rig_utils.lerp(p1[1],p2[1], (float(i)/float(nPnts-1))), arc_rig_utils.lerp(p1[2],p2[2], (float(i)/float(nPnts-1)))]
			self.addLocator("%s_%i"%(prefix,i+1), last_loc, pos)
			last_loc = "%s_%i"%(prefix,i+1)
		
		self.addLocator("%s_%i"%(prefix,nPnts), last_loc, p2)
		self.updateJointsList()
		
		arc_rig_utils.setSelectMode(self.cCreateMultiple_prev_mode)
		cmds.deleteUI(self.cnjl_wnd)
		pass
		
	def _createMultiplePoints_manual_callback(self,a):
		nPnts = cmds.intField(self.cmp_points_2, q=True, v=True)
		p1 = [cmds.floatField(self.cmp_pos1X, q=True, v=True), cmds.floatField(self.cmp_pos1Y, q=True, v=True), cmds.floatField(self.cmp_pos1Z, q=True, v=True)]
		p2 = [cmds.floatField(self.cmp_pos2X, q=True, v=True), cmds.floatField(self.cmp_pos2Y, q=True, v=True), cmds.floatField(self.cmp_pos2Z, q=True, v=True)]
		prefix = cmds.textFieldGrp(self.cmp_name_2, q=True, tx=True)
		parent = cmds.optionMenu(self.cmp_parent2, q=True, v=True)
		
		self.addLocator("%s_1"%prefix, parent, p1)
		last_loc = "%s_1"%prefix
		
		for i in xrange(1,nPnts-1):
			pos = [arc_rig_utils.lerp(p1[0],p2[0], (float(i)/float(nPnts-1))), arc_rig_utils.lerp(p1[1],p2[1], (float(i)/float(nPnts-1))), arc_rig_utils.lerp(p1[2],p2[2], (float(i)/float(nPnts-1)))]
			self.addLocator("%s_%i"%(prefix,i+1), last_loc, pos)
			last_loc = "%s_%i"%(prefix,i+1)
		
		self.addLocator("%s_%i"%(prefix,nPnts), last_loc, p2)
		self.updateJointsList()
		arc_rig_utils.setSelectMode(self.cCreateMultiple_prev_mode)
		cmds.deleteUI(self.cnjl_wnd)
	
	
	def createManualJoint(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
		
		self.cnjl_wnd = cmds.window(title="Create Manual Joint", w=500, s=False)
		form = cmds.formLayout()
		
		ok_btn = cmds.button("ok1", label="OK", command=self._createManualJoint_callback, w=50)
		lout = cmds.text(width=500,label="Enter position then press OK:" ,font="boldLabelFont",ww=True)
		pos1_ff = cmds.text(label="position: " , w=50)
		self.mj_posX = cmds.floatField(pre=4, w=45)
		self.mj_posY = cmds.floatField(pre=4, w=45)
		self.mj_posZ = cmds.floatField(pre=4, w=45)
		
		self.mj_name = cmds.textFieldGrp(l="Name: ", tx="")
		
		self.mj_parent = cmds.optionMenu(l="Parent:	")
		cmds.menuItem(l="-")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.mj_parent, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.mj_name, "top", 10, lout),
		(self.mj_parent, "top", 10, self.mj_name),
		(pos1_ff, "top", 15, self.mj_parent),
		(self.mj_posX, "top", 12, self.mj_parent),(self.mj_posX, "left", 1, pos1_ff),
		(self.mj_posY, "top", 12, self.mj_parent),(self.mj_posY, "left", 1, self.mj_posX),
		(self.mj_posZ, "top", 12, self.mj_parent),(self.mj_posZ, "left", 1, self.mj_posY),
		(ok_btn, "top", 15, pos1_ff),
		]
		
		attachPosition =[
		(self.mj_parent, "left", 0, 13),
		(self.mj_parent, "right", 0, 50),
		(pos1_ff, "left", 0, 28),
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		
		cmds.showWindow(self.cnjl_wnd)
	
	def _createManualJoint_callback(self, a):
		nPos = [cmds.floatField(self.mj_posX, q=True, v=True), cmds.floatField(self.mj_posY, q=True, v=True), cmds.floatField(self.mj_posZ, q=True, v=True)]
		name = cmds.textFieldGrp(self.mj_name, q=True, tx=True)
		parent = cmds.optionMenu(self.mj_parent, q=True, v=True)
		
		foundP = False
		for i in range(len(self.joints)):
			if self.joints[i][0] == name:
				print "Error: Name already exits!"
				return
			if self.joints[i][0] == parent:
				foundP = True
		
		if not foundP:
			if parent != "-":
				print "Error: Parent can't be found!"
				return
		
		self.addLocator(name, parent, nPos)
		self.updateJointsList()
		cmds.deleteUI(self.cnjl_wnd)
		
	def createJointsToPoints(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
				
		self.cJoints2Points_prev_mode = arc_rig_utils.getSelectMode()
		# Choose the vertexs that the joints will be in the middle of
		arc_rig_utils.setSelectMode(arc_rig_utils.SELECTMODE_COM)
		
		self.cnjl_wnd = cmds.window(title="Joints To Points", w=500, s=False)
		form = cmds.formLayout()
		
		ok_btn = cmds.button("ok1", label="OK", command=self._createJointsToPoints_callback, w=50)
		lout = cmds.text(width=500,label="Select the points in order, then press OK" ,font="boldLabelFont",ww=True)
		self.cj2p_name = cmds.textFieldGrp(l="Prefix: ", tx="")
		self.cj2p_parent = cmds.optionMenu(l="Parent:	")
		cmds.menuItem(l="-")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.cj2p_parent, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.cj2p_name, "top", 10, lout),
		(self.cj2p_parent, "top", 10, self.cj2p_name),
		(ok_btn, "top", 15, self.cj2p_parent),
		]
		
		attachPosition =[
		(self.cj2p_parent, "left", 0, 13),
		(self.cj2p_parent, "right", 0, 50),
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		
		cmds.showWindow(self.cnjl_wnd)
	
	def _createJointsToPoints_callback(self, a):
		
		pnts = arc_rig_utils.getSelectedLocations()
		
		if(len(pnts) < 0):
			print "Error: No points selected!"
			return
		
		
		name = cmds.textFieldGrp(self.cj2p_name, q=True, tx=True)
		parent = cmds.optionMenu(self.cj2p_parent, q=True, v=True)
		
		self.addLocator("%s_1"%name, parent, pnts[0])
		lastloc = "%s_1"%name
		
		for i in xrange(1,len(pnts)):
			self.addLocator("%s_%i"%(name,i+1), lastloc, pnts[i])
			lastloc = "%s_%i"%(name,i+1)
		
		self.updateJointsList()
		arc_rig_utils.setSelectMode(self.cJoints2Points_prev_mode)
		cmds.deleteUI(self.cnjl_wnd)
		pass
		
	
	
	def mirrorJoints(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
		
		self.cnjl_wnd = cmds.window(title="Mirror Joints", w=500, s=False)
		form = cmds.formLayout()
		
		
		lout = cmds.text(width=500,label="Select the joint that would start the sequence (Shoulder if you were mirroring an arm)\nThen set the prefix to change (L_Arm to R_Arm)\nThe pivot should usually be where the root joint is located\nThen hit OK" ,font="boldLabelFont",ww=True, align='center')
		
		
		self.mrrj_ori_prefix = cmds.textFieldGrp(l="Original Prefix: ", tx="L_")
		self.mrrj_new_prefix = cmds.textFieldGrp(l="New Prefix:      ", tx="R_")
		
		
		pivot_lbl = cmds.text(label="pivot: " , w=50)
		self.mjjr_pivX = cmds.floatField(pre=4, w=45)
		self.mjjr_pivY = cmds.floatField(pre=4, w=45)
		self.mjjr_pivZ = cmds.floatField(pre=4, w=45)
		
		
		ok_btn = cmds.button("ok1", label="OK", command=self._mirrorJoints_callback, w=50)
		
		self.mrrj_maxis = cmds.optionMenu(l="Mirror Axis:   ")
		for e in ("X", "Y", "Z"):
			cmds.menuItem(l=e)
		
		
		self.mrrj_start = cmds.optionMenu(l="Starting Joint:   ")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.mrrj_start, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.mrrj_ori_prefix, "top", 10, lout),
		(self.mrrj_new_prefix, "top", 10, self.mrrj_ori_prefix),
		(self.mrrj_maxis, "top", 10, self.mrrj_new_prefix),
		(self.mrrj_start, "top", 10, self.mrrj_maxis),
		(pivot_lbl, "top", 10, self.mrrj_start),
		(self.mjjr_pivX, "top", 10, self.mrrj_start), (self.mjjr_pivX, "left", 1, pivot_lbl),
		(self.mjjr_pivY, "top", 10, self.mrrj_start), (self.mjjr_pivY, "left", 1, self.mjjr_pivX),
		(self.mjjr_pivZ, "top", 10, self.mrrj_start), (self.mjjr_pivZ, "left", 1, self.mjjr_pivY),
		(ok_btn, "top", 15, pivot_lbl),
		]
		
		attachPosition =[
		(self.mrrj_maxis, "left", 0, 13),
		(self.mrrj_maxis, "right", 0, 50),
		(self.mrrj_start, "left", 0, 13),
		(self.mrrj_start, "right", 0, 50),
		(pivot_lbl, "left", 0, 28),
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.cnjl_wnd)
		
		
	def _mirrorJoints_callback(self,a):
		
		starting_joint = cmds.optionMenu(self.mrrj_start, q=True, v=True)
		if starting_joint == None or (len(starting_joint) < 1):
			print "Error: No joint selected!"
			return
		
		axis = cmds.optionMenu(self.mrrj_maxis, q=True, v=True)
		ori_prefix = cmds.textFieldGrp(self.mrrj_ori_prefix, q=True, tx=True)
		new_prefix = cmds.textFieldGrp(self.mrrj_new_prefix, q=True, tx=True)
		
		pivot = [cmds.floatField(self.mjjr_pivX, q=True, v=True), cmds.floatField(self.mjjr_pivY, q=True, v=True), cmds.floatField(self.mjjr_pivZ, q=True, v=True)]
		
		
		
		# BUILD CONNECTIONS
		
		# get first new location, along with first parent
		for j in self.joints:
			if j[0] == starting_joint:
				pos = cmds.xform(j[2], q=True, ws=True, a=True, t=True)
				parent = j[1]
				break
		
		if(axis == "X"):
			pos[0] = (pivot[0]-(pos[0]-pivot[0]))
		if(axis == "Y"):
			pos[1] = (pivot[1]-(pos[1]-pivot[1]))
		if(axis == "Z"):
			pos[2] = (pivot[2]-(pos[2]-pivot[2]))
		
		joints_to_affect = [starting_joint]
		jta_pos = [pos]
		jta_parents = [parent]
		jta = 1
		
		for p in range(len(self.joints)):
			o_jta = jta
			for i in range(len(self.joints)):
				
				# Dont want to add duplicates
				if not (self.joints[i][0] in joints_to_affect):
					
					# Add anything that has parent in the current list, to the list
					# Also figure out its position while were here
					# Along with parent
					if (self.joints[i][1] in joints_to_affect):
						joints_to_affect.append(self.joints[i][0])
						pos = cmds.xform(self.joints[i][2], q=True, ws=True, a=True, t=True)
						jta_parents.append(self.joints[i][1])
						
						if(axis == "X"):
							pos[0] = (pivot[0]-(pos[0]-pivot[0]))
							
						if(axis == "Y"):
							pos[1] = (pivot[1]-(pos[1]-pivot[1]))
							
						if(axis == "Z"):
							pos[2] = (pivot[2]-(pos[2]-pivot[2]))
						
						jta_pos.append(pos)
						jta = jta + 1
				
			# No new children found
			if o_jta == jta:
				break
		
		
		
		# ADD THE NEW JOINTS!
		ori_len = len(ori_prefix)
		for i in range(len(joints_to_affect)):
			
			parent = jta_parents[i]
			nJoint = joints_to_affect[i]
			
			# Leave the starting joints parent alone
			if i != 0:
				if parent.startswith(ori_prefix):
					parent = "%s%s"%(new_prefix, parent[ori_len:])
			
			if nJoint.startswith(ori_prefix):
				nJoint = "%s%s"%(new_prefix, nJoint[ori_len:])
			
			self.addLocator(nJoint, parent, jta_pos[i])
		
		
		self.updateJointsList()
		cmds.deleteUI(self.cnjl_wnd)
		
	
	
	def createNewBendyIK(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
			
		if len(self.joints) < 2:
			print "Error: You need atleast 2 joints to create an IK!"
			return
		
		self.cnjl_wnd = cmds.window(title="Create New Bendy IK", w=500, s=False)
		form = cmds.formLayout()
		
		ok_btn = cmds.button("ok1", label="OK", command=self._createNewBendyIK_callback, w=50)
		lout = cmds.text(width=500,label="Create new Bendy IK Handle" ,font="boldLabelFont",ww=True)
		nik_j_lbl = cmds.text(label=" Controllers:")
		self.nik_jnts = cmds.intField("nik_jnts", min=4, v=4)
		self.nik_ctrlsz = cmds.floatFieldGrp(label="Controller Size: ", value1=1.0)
		self.nik_name = cmds.textFieldGrp(l="Name: ", tx="")
		self.nik_startj = cmds.optionMenu(l="Starting Joint: ")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		
		self.nik_endj = cmds.optionMenu(l="Ending Joint:   ")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.nik_endj, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.nik_name, "top", 10, lout),
		(self.nik_startj, "top", 10, self.nik_name),
		(self.nik_endj, "top", 10, self.nik_startj),
		(self.nik_jnts, "top", 17, self.nik_endj), 
		(nik_j_lbl, "top", 20, self.nik_endj), (nik_j_lbl, "left", 5, self.nik_jnts),
		(self.nik_ctrlsz, "top", 10, nik_j_lbl),
		(ok_btn, "top", 15, self.nik_ctrlsz),
		]
		
		attachPosition =[
		(self.nik_startj, "left", 0, 13),
		(self.nik_startj, "right", 0, 50),
		(self.nik_endj, "left", 0, 13),
		(self.nik_endj, "right", 0, 50),
		(self.nik_jnts, "left", 0, 30),
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		
		cmds.showWindow(self.cnjl_wnd)
		
		
		
	def _createNewBendyIK_callback(self, a):
		
		name = cmds.textFieldGrp(self.nik_name, q=True, tx=True)
		startj = cmds.optionMenu(self.nik_startj, q=True, v=True)
		endj = cmds.optionMenu(self.nik_endj, q=True, v=True)
		numberOfJoints = cmds.intField(self.nik_jnts, q=True, v=True)-3
		csize = cmds.floatFieldGrp(self.nik_ctrlsz , q=True , v1=True)
		
		self.bendy_iks.append([name,startj,endj,numberOfJoints,csize])
		
		
		self.updateIKsList()
		cmds.deleteUI(self.cnjl_wnd)
		
		
		
	def createNewPVIK(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
			
		if len(self.joints) < 2:
			print "Error: You need atleast 2 joints to create an IK!"
			return
		
		self.cnjl_wnd = cmds.window(title="Create New Normal IK", w=500, s=False)
		form = cmds.formLayout()
		
		ok_btn = cmds.button("ok1", label="OK", command=self._createNewPVIK_callback, w=50)
		lout = cmds.text(width=500,label="Create new Normal IK Handle" ,font="boldLabelFont",ww=True)
		
		self.pvik_lo = cmds.floatFieldGrp( numberOfFields=3, label="Pole Vector Offset:", value1=0.0, value2=-1.0, value3=-3.0 )
		self.pvik_ctrlsz = cmds.floatFieldGrp(label="Controller Size:", value1=1.0)
		self.pvik_locatorsz = cmds.floatFieldGrp(label="Locator Size:", value1=0.5)
		self.pvik_name = cmds.textFieldGrp(l="Name: ", tx="")
		
		self.pvik_startj = cmds.optionMenu(l="Starting Joint: ")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		
		self.pvik_endj = cmds.optionMenu(l="Ending Joint:   ")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.pvik_endj, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.pvik_name, "top", 10, lout),
		(self.pvik_startj, "top", 10, self.pvik_name),
		(self.pvik_endj, "top", 10, self.pvik_startj),
		(self.pvik_ctrlsz, "top", 17, self.pvik_endj), 
		(self.pvik_locatorsz, "top", 10, self.pvik_ctrlsz),
		(self.pvik_lo, "top", 10, self.pvik_locatorsz),
		(ok_btn, "top", 15, self.pvik_lo),
		]
		
		attachPosition =[
		(self.pvik_startj, "left", 0, 13),
		(self.pvik_startj, "right", 0, 50),
		(self.pvik_endj, "left", 0, 13),
		(self.pvik_endj, "right", 0, 50),
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		
		cmds.showWindow(self.cnjl_wnd)
		
	def _createNewPVIK_callback(self, a):
		
		name = cmds.textFieldGrp(self.pvik_name, q=True, tx=True)
		startj = cmds.optionMenu(self.pvik_startj, q=True, v=True)
		endj = cmds.optionMenu(self.pvik_endj, q=True, v=True)
		csize = cmds.floatFieldGrp(self.pvik_ctrlsz , q=True , v1=True)
		lsize = cmds.floatFieldGrp(self.pvik_locatorsz , q=True , v1=True)
		lo = cmds.floatFieldGrp(self.pvik_lo , q=True , v=True)
		
		self.pv_iks.append([name,startj,endj,csize,lsize,lo[0],lo[1],lo[2]])
		
		self.updateIKsList()
		cmds.deleteUI(self.cnjl_wnd)
	
	
	def updateIKsList(self):
	
		# Cant update something that's not there
		if not cmds.window(self.editik_wnd, q=True, ex=True):
			return
		
		l = len(self.bendy_iks)
		
		self.ignoreIkUpdates = True
		
		try:
			# Doing this stops scriptTable freaking out whenever someone changes a value and then adding a new ik
			# It seems to be because it tries to force a value on whatevers selectd, doing this removes everything, stopping that
			cmds.scriptTable(self.bendy_ikstable, e=True, r=0)
			cmds.scriptTable(self.bendy_ikstable, e=True, r=l)
			for i in range(l):
				cmds.scriptTable(self.bendy_ikstable, e=True, ci=[i+1,1], cv=self.bendy_iks[i][0]) 
				cmds.scriptTable(self.bendy_ikstable, e=True, ci=[i+1,2], cv=self.bendy_iks[i][1])
				cmds.scriptTable(self.bendy_ikstable, e=True, ci=[i+1,3], cv=self.bendy_iks[i][2])
				cmds.scriptTable(self.bendy_ikstable, e=True, ci=[i+1,4], cv=self.bendy_iks[i][3]+3)
				cmds.scriptTable(self.bendy_ikstable, e=True, ci=[i+1,5], cv=self.bendy_iks[i][4])
		except Exception as e:
			pass
			
			
			
		l = len(self.pv_iks)
			
		try:
			# Doing this stops scriptTable freaking out whenever someone changes a value and then adding a new ik
			# It seems to be because it tries to force a value on whatevers selectd, doing this removes everything, stopping that
			cmds.scriptTable(self.pv_ikstable, e=True, r=0)
			cmds.scriptTable(self.pv_ikstable, e=True, r=l)
			for i in range(l):
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,1], cv=self.pv_iks[i][0]) 
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,2], cv=self.pv_iks[i][1])
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,3], cv=self.pv_iks[i][2])
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,4], cv=self.pv_iks[i][3])
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,5], cv=self.pv_iks[i][4])
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,6], cv=self.pv_iks[i][5])
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,7], cv=self.pv_iks[i][6])
				cmds.scriptTable(self.pv_ikstable, e=True, ci=[i+1,8], cv=self.pv_iks[i][7])
		except Exception as e:
			print e
			pass
		
		self.ignoreIkUpdates = False
	
	
	def updateBendyIKTableCCC(self, r,c,v):
		if self.ignoreIkUpdates == True:
			return 1
		
		
		# Name changing
		if c == 1:
			for i in range(len(self.bendy_iks)):
				if(self.bendy_iks[i][0] == v):
					print "Error: Theres already something named that!"
					return 0
			self.bendy_iks[r-1][0] = v
			return 1
			
		# Start joint changing
		if c == 2:
			for i in range(len(self.joints)):
				if(self.joints[i][0] == v):
					self.bendy_iks[r-1][1] = v
					return 1
			print "Error: Joint doesn't exist!"
			return 0
			
		# End joint changing
		if c == 3:
			for i in range(len(self.joints)):
				if(self.joints[i][0] == v):
					self.bendy_iks[r-1][2] = v
					return 1
			print "Error: Joint doesn't exist!"
			return 0
		
		# Number of joints
		
		if c == 4:
			try:
				v = int(v)
			except:
				print "Error: Not an integer!"
				return 0
			
			if v < 4:
				print "Error: Must be 4 or greater!"
				return 0
			
			
			self.bendy_iks[r-1][3] = v-3
			return 1
		
		# Control Size
		if c == 5:
			
			try:
				v = float(v)
			except:
				print "Error: Not a float!"
				return 0
			
			self.bendy_iks[r-1][4] = v
		
		
		return 1
		
	
	def updatePVIKTableCCC(self, r,c,v):
		if self.ignoreIkUpdates == True:
			return 1
		
		# Name changing
		if c == 1:
			for i in range(len(self.pv_iks)):
				if(self.pv_iks[i][0] == v):
					print "Error: Theres already something named that!"
					return 0
			self.pv_iks[r-1][0] = v
			return 1
			
		# Start joint changing
		if c == 2:
			for i in range(len(self.joints)):
				if(self.joints[i][0] == v):
					self.pv_iks[r-1][1] = v
					return 1
			print "Error: Joint doesn't exist!"
			return 0
			
		# End joint changing
		if c == 3:
			for i in range(len(self.joints)):
				if(self.joints[i][0] == v):
					self.pv_iks[r-1][2] = v
					return 1
			print "Error: Joint doesn't exist!"
			return 0
		
		
		# Control Size
		if c == 4:
			
			try:
				v = float(v)
			except:
				print "Error: Not a float!"
				return 0
			
			self.pv_iks[r-1][3] = v
			return 1
		
		
		# Locator Size
		if c == 5:
			
			try:
				v = float(v)
			except:
				print "Error: Not a float!"
				return 0
			
			self.pv_iks[r-1][4] = v
			return 1
			
			
		# Pole Vector Offset
		try:
			v = float(v)
		except:
			print "Error: Not a float!"
			return 0
		
		self.pv_iks[r-1][c-1] = v
		
		return 1
	
	def editIKs(self, a):
		# duplicates are not ok
		if cmds.window(self.editik_wnd, q=True, ex=True):
			return
		
		self.editik_wnd = cmds.window(title="Edit IK Handles", w=600)
		form = cmds.formLayout()
		
		self.bendy_ik_txt = cmds.text(width=500,label="Bendy IKs" ,font="boldLabelFont",ww=True)
		self.pv_ik_txt = cmds.text(width=500,label="Normal IKs" ,font="boldLabelFont",ww=True)
		
		self.bendy_ikstable = cmds.scriptTable(h=200, rows=0,columns=5, l=[(1,"name"),(2,"start joint"), (3,"end joint"), (4,"controls"),(5,"control size")], cw=[(1,120),(2,120),(3,120), (4,70), (5,120)], ccc=self.updateBendyIKTableCCC)
		
		self.pv_ikstable = cmds.scriptTable(h=200,rows=0,columns=8, l=[(1,"name"),(2,"start joint"), (3,"end joint"), (4,"control size"),(5,"locator size"),(6,"PVO x"), (7,"PVO y"), (8,"PVO z")], cw=[(1,120),(2,120),(3,120), (4,65), (5,65), (6,45), (7,45), (8,45)], ccc=self.updatePVIKTableCCC)
		
		self.ik_deleteTxt = cmds.textFieldGrp(l="Delete IK: ", tx="")
		self.ik_deleteBtn = cmds.button("del1", label="Delete", command=self._editIKs_delete, w=50)
		
		self.updateIKsList()
		
		
		
		
		attachForm = [
		(self.bendy_ik_txt, "left", 10),(self.bendy_ik_txt, "right", 10),(self.bendy_ik_txt,"top",10),
		(self.bendy_ikstable, "left", 10),(self.bendy_ikstable, "right", 10),
		(self.pv_ik_txt, "left", 10),(self.pv_ik_txt, "right", 10),
		(self.pv_ikstable, "left", 10),(self.pv_ikstable, "right", 10),(self.pv_ikstable,"bottom",40),
		]
		
		attachControl =[
		(self.bendy_ikstable, "top", 10, self.bendy_ik_txt),
		(self.pv_ik_txt, "top", 20, self.bendy_ikstable),
		(self.pv_ikstable, "top", 10, self.pv_ik_txt),
		(self.ik_deleteTxt, "top", 10, self.pv_ikstable),
		(self.ik_deleteBtn, "top", 10, self.pv_ikstable), (self.ik_deleteBtn, "left", 5, self.ik_deleteTxt),
		]
		
		attachPosition =[
		(self.ik_deleteTxt, "left", 0, 0),
		]
		
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.editik_wnd)
	
	
	def _editIKs_delete(self, a):
		name = cmds.textFieldGrp(self.ik_deleteTxt, q=True, tx=True)
		
		try:
			for i in range(len(self.bendy_iks)):
				if(self.bendy_iks[i][0] == name):
					self.bendy_iks.pop(i)
					self.updateIKsList()
					cmds.textFieldGrp(self.ik_deleteTxt, e=True, tx="")
					return
		except:
			pass
			
			
		try:
			for i in range(len(self.pv_iks)):
				if(self.pv_iks[i][0] == name):
					self.pv_iks.pop(i)
					self.updateIKsList()
					cmds.textFieldGrp(self.ik_deleteTxt, e=True, tx="")
					return
		except:
			pass
		
		print "Error: Couldn't delete IK!"
		
	
	def createNewAimConstraint(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
			
		if len(self.joints) < 1:
			print "Error: You need atleast 1 joints to create an Geo deformer!"
			return
		
		self.cnewaim_con_prev_mode = arc_rig_utils.getSelectMode()
		arc_rig_utils.setSelectMode(arc_rig_utils.SELECTMODE_OBJ)
		
		self.cnjl_wnd = cmds.window(title="New Eye Constraint", w=500, s=False)
		form = cmds.formLayout()
		
		ok_btn = cmds.button("ok1", label="OK", command=self._createNewAimConstraint_callback, w=50)
		lout = cmds.text(width=500,label="Select the Object you wish to constraint, then press OK." ,font="boldLabelFont",ww=True)
		
		
		self.cnac_name = cmds.textFieldGrp(l="Name: ", tx="")
		self.cnac_ctrlsz = cmds.floatFieldGrp(label="Controller Size:", value1=0.1)
		self.cnac_co = cmds.floatFieldGrp( numberOfFields=3, label="Controller Offset:", value1=0.0, value2=0.0, value3=2.0 )
		
		
		self.cnac_parent = cmds.optionMenu(l="Parent:	")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.cnac_parent, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.cnac_name, "top", 10, lout),
		(self.cnac_parent, "top", 10, self.cnac_name),
		(self.cnac_ctrlsz, "top", 10, self.cnac_parent),
		(self.cnac_co, "top", 10, self.cnac_ctrlsz),
		(ok_btn, "top", 15, self.cnac_co),
		]
		
		attachPosition =[
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.cnjl_wnd)
		
		
	def _createNewAimConstraint_callback(self, a):
		
		try:
			parent = cmds.ls(selection=True)[0]
		except:
			print "Error: Select an Object first!"
			return
			
		
		name = cmds.textFieldGrp(self.cnac_name, q=True, tx=True)
		csize = cmds.floatFieldGrp(self.cnac_ctrlsz , q=True , v1=True)
		co = cmds.floatFieldGrp(self.cnac_co , q=True , v=True)
		parent_joint = cmds.optionMenu(self.cnac_parent, q=True, v=True)
		
		
		self.geo_aimc.append([name,parent,csize,co[0],co[1],co[2],parent_joint])
		
		self.updateGeoDefsList()
		arc_rig_utils.setSelectMode(self.cnewaim_con_prev_mode)
		cmds.deleteUI(self.cnjl_wnd)
		
		
		
	def createNewPointController(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
		
		if len(self.joints) < 1:
			print "Error: You need atleast 1 joints to create an Geo deformer!"
			return
		
		self.cnewpc_prev_mode = arc_rig_utils.getSelectMode()
		arc_rig_utils.setSelectMode(arc_rig_utils.SELECTMODE_COM)
		
		self.cnjl_wnd = cmds.window(title="New Point Controller", w=500, s=False)
		form = cmds.formLayout()
		
		ok_btn = cmds.button("ok1", label="OK", command=self._createNewPointController_callback, w=50)
		lout = cmds.text(width=500,label="Select the Point(s) you wish to create controllers for\nThen press OK." ,font="boldLabelFont",ww=True)
		
		
		self.cnpc_name = cmds.textFieldGrp(l="Name: ", tx="")
		self.cnpc_ctrlsz = cmds.floatFieldGrp(label="Controller Size:", pre=4, value1=0.05)
		self.cnpc_dst = cmds.floatFieldGrp(label="Distance:", pre=4, value1=0.2)
		
		self.cnpc_parent = cmds.optionMenu(l="Parent:	")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.cnpc_parent, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.cnpc_name, "top", 10, lout),
		(self.cnpc_parent, "top", 10, self.cnpc_name),
		(self.cnpc_ctrlsz, "top", 10, self.cnpc_parent),
		(self.cnpc_dst, "top", 10, self.cnpc_ctrlsz),
		(ok_btn, "top", 15, self.cnpc_dst),
		]
		
		attachPosition =[
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.cnjl_wnd)
		
		
	def _createNewPointController_callback(self, a):
		
		try:
			parent_points = cmds.ls(fl=1,os=1)
			if len(parent_points) < 1:
				raise Exception("")
		except:
			print "Error: Select atleast one Point first!"
			return
			
		
		name = cmds.textFieldGrp(self.cnpc_name, q=True, tx=True)
		csize = cmds.floatFieldGrp(self.cnpc_ctrlsz , q=True , v1=True)
		dst = cmds.floatFieldGrp(self.cnpc_dst , q=True , v1=True)
		parent_joint = cmds.optionMenu(self.cnpc_parent, q=True, v=True)
		
		if len(parent_points) > 1:			
			for i in range(len(parent_points)):
				self.geo_pc.append(["%s_%i"%(name,i),parent_points[i],csize,dst,parent_joint])
			
		else:
			print parent_points[0]
			self.geo_pc.append([name,parent_points[0],csize,dst,parent_joint])
		
		
		self.updateGeoDefsList()
		arc_rig_utils.setSelectMode(self.cnewpc_prev_mode)
		cmds.deleteUI(self.cnjl_wnd)
		
	
	def createNewSSPointController(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
		
		if len(self.joints) < 1:
			print "Error: You need atleast 1 joints to create an Geo deformer!"
			return
		
		self.cnewsspc_prev_mode = arc_rig_utils.getSelectMode()
		arc_rig_utils.setSelectMode(arc_rig_utils.SELECTMODE_COM)
		
		self.cnjl_wnd = cmds.window(title="New SS Point Controller", w=500, s=False)
		form = cmds.formLayout()
		
		ok_btn = cmds.button("ok1", label="OK", command=self._createNewSSPointController_callback, w=50)
		lout = cmds.text(width=500,label="Select the Point(s) you wish to create controllers for\nThen press OK." ,font="boldLabelFont",ww=True)
		
		
		self.cnsspc_name = cmds.textFieldGrp(l="Name: ", tx="")
		self.cnsspc_ctrlsz = cmds.floatFieldGrp(label="Controller Size:", pre=4, value1=1)
		self.cnsspc_dst = cmds.floatFieldGrp(label="Distance:", pre=4, value1=0.5)
		self.cnsspc_fr = cmds.floatFieldGrp(label="Falloff Radius:", pre=4, value1=1)
		
		
		self.cnsspc_parent = cmds.optionMenu(l="Parent:	")
		for e in self.joints:
			cmds.menuItem(l=e[0])
		try:
			cmds.optionMenu(self.cnsspc_parent, e=True, v=self.joints[len(self.joints)-1][0])
		except:
			pass
		
		attachForm = [
		(lout, "left", 0),(lout, "right", 0),(lout,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(self.cnsspc_name, "top", 10, lout),
		(self.cnsspc_parent, "top", 10, self.cnsspc_name),
		(self.cnsspc_ctrlsz, "top", 10, self.cnsspc_parent),
		(self.cnsspc_dst, "top", 10, self.cnsspc_ctrlsz),
		(self.cnsspc_fr, "top", 10, self.cnsspc_dst),
		(ok_btn, "top", 15, self.cnsspc_fr),
		]
		
		attachPosition =[
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.cnjl_wnd)
		
		
	def _createNewSSPointController_callback(self, a):
		
		try:
			parent_points = cmds.ls(fl=1,os=1)
			if len(parent_points) < 1:
				raise Exception("")
		except:
			print "Error: Select atleast one Point first!"
			return
			
		
		name = cmds.textFieldGrp(self.cnsspc_name, q=True, tx=True)
		csize = cmds.floatFieldGrp(self.cnsspc_ctrlsz , q=True , v1=True)
		dst = cmds.floatFieldGrp(self.cnsspc_dst , q=True , v1=True)
		fr = cmds.floatFieldGrp(self.cnsspc_fr , q=True , v1=True)
		parent_joint = cmds.optionMenu(self.cnsspc_parent, q=True, v=True)
		
		if len(parent_points) > 1:			
			for i in range(len(parent_points)):
				self.geo_sspc.append(["%s_%i"%(name,i),parent_points[i],csize,dst,fr,parent_joint])
			
		else:
			self.geo_sspc.append([name,parent_points[0],csize,dst,fr,parent_joint])
		
		
		self.updateGeoDefsList()
		arc_rig_utils.setSelectMode(self.cnewsspc_prev_mode)
		cmds.deleteUI(self.cnjl_wnd)
	
	
	
	
	def updateACTAbleCCCself(self, r,c,v):
		if self.ignoreGeoDUpdates == True:
			return 1
		
		# Name changing
		if c == 1:
			for i in range(len(self.geo_aimc)):
				if(self.geo_aimc[i][0] == v):
					print "Error: Theres already something named that!"
					return 0
			self.geo_aimc[r-1][0] = v
			return 1
			
			
		
		# Parent Object
		if c == 2:
			
			try:
				sel = cmds.select(v)
			except:
				print "Error: cannot find object!"
				return 0
			
			self.geo_aimc[r-1][1] = v
			return 1
		
		
		# Parent Joint
		if c == 7:
			for i in range(len(self.joints)):
				if(self.joints[i][0] == v):
					self.geo_aimc[r-1][6] = v
					return 1
			print "Error: Joint doesn't exist!"
			return 0
			
			
			
		# Everything else is a float
		try:
			v = float(v)
		except:
			print "Error: Not a float!"
			return 0
		
		self.geo_aimc[r-1][c-1] = v
		
		
		return 1
	
	def updatePCTAbleCCCself(self, r,c,v):
		if self.ignoreGeoDUpdates == True:
			return 1
			
		# Name changing
		if c == 1:
			for i in range(len(self.geo_pc)):
				if(self.geo_pc[i][0] == v):
					print "Error: Theres already something named that!"
					return 0
			self.geo_pc[r-1][0] = v
			return 1
			
		
		# Parent Point
		if c == 2:
			
			try:
				sel = cmds.select(v)
			except:
				print "Error: cannot find point!"
				return 0
			
			self.geo_pc[r-1][1] = v
			return 1
			
			
		
		# Parent Joint
		if c == 5:
			for i in range(len(self.joints)):
				if(self.joints[i][0] == v):
					self.geo_pc[r-1][4] = v
					return 1
			print "Error: Joint doesn't exist!"
			return 0
			
			
		
		# Everything else is a float
		try:
			v = float(v)
		except:
			print "Error: Not a float!"
			return 0
		
		self.geo_pc[r-1][c-1] = v
		
			
		return 1
	
	
	def updateSSPCTAbleCCCself(self, r,c,v):
		if self.ignoreGeoDUpdates == True:
			return 1
		
		# Name changing
		if c == 1:
			for i in range(len(self.geo_sspc)):
				if(self.geo_sspc[i][0] == v):
					print "Error: Theres already something named that!"
					return 0
			self.geo_sspc[r-1][0] = v
			return 1
			
		
		# Parent Point
		if c == 2:
			
			try:
				sel = cmds.select(v)
			except:
				print "Error: cannot find point!"
				return 0
			
			self.geo_sspc[r-1][1] = v
			return 1
			
			
		
		# Parent Joint
		if c == 6:
			for i in range(len(self.joints)):
				if(self.joints[i][0] == v):
					self.geo_sspc[r-1][5] = v
					return 1
			print "Error: Joint doesn't exist!"
			return 0
			
			
		# Everything else is a float
		try:
			v = float(v)
		except:
			print "Error: Not a float!"
			return 0
		
		self.geo_sspc[r-1][c-1] = v
			
			
		return 1
	
	
	def updateGeoDefsList(self):
		# Cant update something that's not there
		if not cmds.window(self.editdef_wnd, q=True, ex=True):
			return
		
		self.ignoreGeoDUpdates = True
		
		l = len(self.geo_aimc)
		try:
			# Doing this stops scriptTable freaking out whenever someone changes a value and then adding a new ik
			# It seems to be because it tries to force a value on whatevers selectd, doing this removes everything, stopping that
			cmds.scriptTable(self.ac_geotable, e=True, r=0)
			cmds.scriptTable(self.ac_geotable, e=True, r=l)
			for i in range(l):
				for j in range(len(self.geo_aimc[0])):
					cmds.scriptTable(self.ac_geotable, e=True, ci=[i+1,j+1], cv=self.geo_aimc[i][j])
		except Exception as e:
			pass
			
			
		l = len(self.geo_pc)
		try:
			# Doing this stops scriptTable freaking out whenever someone changes a value and then adding a new ik
			# It seems to be because it tries to force a value on whatevers selectd, doing this removes everything, stopping that
			cmds.scriptTable(self.pc_geotable, e=True, r=0)
			cmds.scriptTable(self.pc_geotable, e=True, r=l)
			for i in range(l):
				for j in range(len(self.geo_pc[0])):
					cmds.scriptTable(self.pc_geotable, e=True, ci=[i+1,j+1], cv=self.geo_pc[i][j])
		except Exception as e:
			pass
			
			
		l = len(self.geo_sspc)
		try:
			# Doing this stops scriptTable freaking out whenever someone changes a value and then adding a new ik
			# It seems to be because it tries to force a value on whatevers selectd, doing this removes everything, stopping that
			cmds.scriptTable(self.sspc_geotable, e=True, r=0)
			cmds.scriptTable(self.sspc_geotable, e=True, r=l)
			for i in range(l):
				for j in range(len(self.geo_sspc[0])):
					cmds.scriptTable(self.sspc_geotable, e=True, ci=[i+1,j+1], cv=self.geo_sspc[i][j])
		except Exception as e:
			pass
		
		self.ignoreGeoDUpdates = False
		
		
	def editDeformers(self, a):
		# duplicates are not ok
		if cmds.window(self.editdef_wnd, q=True, ex=True):
			return
		
		self.editdef_wnd = cmds.window(title="Edit Deformers", w=655)
		form = cmds.formLayout()
		
		self.ac_geo_txt = cmds.text(width=500,label="Eye Aimer" ,font="boldLabelFont",ww=True)
		self.pc_geo_txt = cmds.text(width=500,label="Point Controller" ,font="boldLabelFont",ww=True)
		self.sspc_geo_txt = cmds.text(width=500,label="Softselect Point Controller" ,font="boldLabelFont",ww=True)
		
		self.ac_geotable = cmds.scriptTable(h=200, rows=0,columns=7, l=[(1,"name"),(2,"parent object"),(3, "controller size"), (4,"CO x"), (5,"CO y"),(6,"CO z"), (7,"parent joint")], cw=[(1,120),(2,120),(3,70), (4,70), (5,70), (6,70), (7,120)], ccc=self.updateACTAbleCCCself)

		self.pc_geotable = cmds.scriptTable(h=200,rows=0,columns=5, l=[(1,"name"),(2,"parent point"), (3,"controller size"), (4,"distance"),(5,"parent joint")], cw=[(1,120),(2,120),(3,70), (4,70), (5,120)], ccc=self.updatePCTAbleCCCself)
		
		self.sspc_geotable = cmds.scriptTable(h=200,rows=0,columns=6, l=[(1,"name"),(2,"parent point"), (3,"controller size"), (4,"distance"),(5,"fallout radius"),(6,"parent joint")], cw=[(1,120),(2,120),(3,70), (4,70), (5,70), (6,120)], ccc=self.updateSSPCTAbleCCCself)
		
		self.geo_deleteTxt = cmds.textFieldGrp(l="Delete Deformer: ", tx="")
		self.geo_deleteBtn = cmds.button("del1", label="Delete", command=self._editDeformer_delete, w=50)
		
		self.updateGeoDefsList()
		
		
		
		
		attachForm = [
		(self.ac_geo_txt, "left", 10),(self.ac_geo_txt, "right", 10),(self.ac_geo_txt,"top",10),
		(self.ac_geotable, "left", 10),(self.ac_geotable, "right", 10),
		(self.pc_geo_txt, "left", 10),(self.pc_geo_txt, "right", 10),
		(self.pc_geotable, "left", 10),(self.pc_geotable, "right", 10),
		(self.sspc_geo_txt, "left", 10),(self.sspc_geo_txt, "right", 10),
		(self.sspc_geotable, "left", 10),(self.sspc_geotable, "right", 10), (self.sspc_geotable, "bottom", 40),
		]
		
		attachControl =[
		(self.ac_geotable, "top", 10, self.ac_geo_txt),
		(self.pc_geo_txt, "top", 20, self.ac_geotable),
		(self.pc_geotable, "top", 10, self.pc_geo_txt),
		(self.sspc_geo_txt, "top", 20, self.pc_geotable),
		(self.sspc_geotable, "top", 10, self.sspc_geo_txt),
		(self.geo_deleteTxt, "top", 10, self.sspc_geotable),
		(self.geo_deleteBtn, "top", 10, self.sspc_geotable), (self.geo_deleteBtn, "left", 5, self.geo_deleteTxt),
		]
		
		attachPosition =[
		(self.geo_deleteTxt, "left", 0, 0),
		]
		
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.editdef_wnd)
		
		
	def _editDeformer_delete(self, a):
		
		name = cmds.textFieldGrp(self.geo_deleteTxt, q=True, tx=True)
		
		try:
			for i in range(len(self.geo_aimc)):
				if(self.geo_aimc[i][0] == name):
					self.geo_aimc.pop(i)
					self.updateGeoDefsList()
					cmds.textFieldGrp(self.geo_deleteTxt, e=True, tx="")
					return
		except:
			pass
			
			
		try:
			for i in range(len(self.pc_aimc)):
				if(self.geo_pc[i][0] == name):
					self.geo_pc.pop(i)
					self.updateGeoDefsList()
					cmds.textFieldGrp(self.geo_deleteTxt, e=True, tx="")
					return
		except:
			pass
			
			
		try:
			for i in range(len(self.sspc_aimc)):
				if(self.geo_sspc[i][0] == name):
					self.geo_sspc.pop(i)
					self.updateGeoDefsList()
					cmds.textFieldGrp(self.geo_deleteTxt, e=True, tx="")
					return
		except:
			pass
		
		print "Error: Couldn't delete IK!"
	
	
	def deletePreviewRig(self, *args):
		try:
			cmds.delete("ARCBUILDER__PRERIG__*")
		except:
			pass
		
		
		# KILL PRERIG SCRIPT JOBS!
		conds = cmds.scriptJob( listJobs=True )
		for cond in conds:
			# Check for any job with our prefix in it
			if "ARCBUILDER__PRERIG__" in cond:
				for i in range(len(cond)):
					
					# Extract the id
					if cond[i] == ":":
						break
				
				# Kill the job
				#print cond
				try:
					cmds.scriptJob(kill=int(cond[:i]))
				except:
					pass
		
	
	def previewRig(self,*args):
		self.deletePreviewRig()
		
		
		# Create Groups
		#grp_master = cmds.group(n="ARCBUILDER__PRERIG___GRP", empty=True)
		grp_master = arc_rig_utils.makeGlobalControllerObject("ARCBUILDER__PRERIG___RIG", size=5)[0]
		
		
		if len(self.joints) > 0:
			grp_joints = cmds.group(n="ARCBUILDER__PRERIG__Joints_GRP", empty=True)
			cmds.parent(grp_joints, grp_master)
			lockList = ["%s.%s"%(grp_joints,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
			for v in lockList:
				arc_rig_utils.lockAttr(v)
				arc_rig_utils.hideAttr(v)
			
		
		
		if len(self.bendy_iks) > 0:
			grp_bendy_iks = cmds.group(n="ARCBUILDER__PRERIG__BendyIKs_GRP", empty=True)
			cmds.parent(grp_bendy_iks, grp_master)
			lockList = ["%s.%s"%(grp_bendy_iks,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
			for v in lockList:
				arc_rig_utils.lockAttr(v)
				arc_rig_utils.hideAttr(v)
		
		
		
		if len(self.pv_iks) > 0:
			grp_pv_iks = cmds.group(n="ARCBUILDER__PRERIG__NormalIKs_GRP", empty=True)
			cmds.parent(grp_pv_iks, grp_master)
			lockList = ["%s.%s"%(grp_pv_iks,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
			for v in lockList:
				arc_rig_utils.lockAttr(v)
				arc_rig_utils.hideAttr(v)
		
		
		
		if len(self.geo_aimc)+len(self.geo_pc)+len(self.geo_sspc) > 0:
			grp_dgeo = cmds.group(n="ARCBUILDER__PRERIG__DeformGeos_GRP", empty=True)
			cmds.parent(grp_dgeo, grp_master)
			lockList = ["%s.%s"%(grp_dgeo,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
			for v in lockList:
				arc_rig_utils.lockAttr(v)
				arc_rig_utils.hideAttr(v)
		
		
		
		#Lock /Hide Master Groups transforms
		#lockList = ["%s.%s"%(grp_master,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		#for v in lockList:
		#	arc_rig_utils.lockAttr(v)
		#	arc_rig_utils.hideAttr(v)
		#
		
		
			
		
		joints_cpy = self.joints[:]
		l = len(joints_cpy)
		r_rig_data = [[],[]]
		
		# First add roots, then remove them
		for i in range(l):
			if i >= l:
				break
			if joints_cpy[i][1] == "-":
				r_rig_data[0].append(joints_cpy[i][0])
				r_rig_data[1].append(arc_rig_utils.joint(name=joints_cpy[i][0], prefix="ARCBUILDER__PRERIG__", p=cmds.xform(joints_cpy[i][2], q=True, ws=True, a=True, t=True)))
				
				# Add all roots to master joint group
				cmds.parent(r_rig_data[1][len(r_rig_data[1])-1], grp_joints)
				
				joints_cpy.pop(i)
				i = i-1
				l = l-1
		
		
		# Scroll through the joints, not adding them until a parent is found
		i=0
		while len(joints_cpy) != 0:
			
			parent=joints_cpy[i][1]
			for j in range(len(r_rig_data[0])):
				#print j
				if parent == r_rig_data[0][j]:
					r_rig_data[0].append(joints_cpy[i][0])
					r_rig_data[1].append(arc_rig_utils.joint(owner=r_rig_data[1][j], name=joints_cpy[i][0], prefix="ARCBUILDER__PRERIG__", p=cmds.xform(joints_cpy[i][2], q=True, ws=True, a=True, t=True)))
					joints_cpy.pop(i)
					i = i-1
					break
				
			i = i +1
			if i >= len(joints_cpy):
				i=0
		
		
		
		
		# Add Geo Deformers
		# We're doing this before IKs to stop eye constraints going weird when we skin things
		for geo in self.geo_aimc:
			parent_j = r_rig_data[1][ r_rig_data[0].index(geo[6])]
			aimc = arc_rig_utils.createAimConstraint("ARCBUILDER__PRERIG__%s"%geo[0], geo[1], geo[2], [geo[3],geo[4], geo[5]], parent_j)
			
				
				
		for geo in self.geo_pc:
			parent_j = r_rig_data[1][ r_rig_data[0].index(geo[4])]
			pc = arc_rig_utils.createPointController("ARCBUILDER__PRERIG__%s"%geo[0], geo[1], geo[2], geo[3], parent_j)
			# Group only cluster
			cmds.parent(pc[1][1], grp_dgeo)
				
		for geo in self.geo_sspc:
			parent_j = r_rig_data[1][ r_rig_data[0].index(geo[5])]
			sspc = arc_rig_utils.createSoftPointController("ARCBUILDER__PRERIG__%s"%geo[0], geo[1], geo[2], geo[3], geo[4], parent_j)
			cmds.parent(sspc[2],  parent_j)
		
		
		
		# Add IK Handles
		# We're setting owner to master, so all rig controlls appear there
		
		for ik in self.bendy_iks:
			startj = r_rig_data[1][ r_rig_data[0].index(ik[1]) ]
			endj = r_rig_data[1][ r_rig_data[0].index(ik[2]) ]
			bik = arc_rig_utils.spline_ikhandle("ARCBUILDER__PRERIG__%s"%ik[0], startj, endj, ik[3], controllerSize=ik[4], owner=grp_master)
			cmds.parent(bik[1], grp_bendy_iks)
			
		for ik in self.pv_iks:
			startj = r_rig_data[1][ r_rig_data[0].index(ik[1]) ]
			endj = r_rig_data[1][ r_rig_data[0].index(ik[2]) ]
			pvik = arc_rig_utils.pv_ikhandle("ARCBUILDER__PRERIG__%s"%ik[0], startj, endj, ik[3], ik[4], [ik[5],ik[6],ik[7]], owner=grp_master)
			cmds.parent(pvik[1], grp_pv_iks)
			
			
		
		
	def bakeRig(self, a):
		# duplicates are not ok
		if cmds.window(self.cnjl_wnd, q=True, ex=True):
			return
		
		self.cnjl_wnd = cmds.window(title="Bake Rig", w=500, s=False)
		form = cmds.formLayout()
		
		self.br_rp = cmds.textFieldGrp( label='Rig Prefix', text='' )
		ok_btn = cmds.button(label="Bake Rig", align="center", w=50, command=self._bakeRig_callback)
		
		attachForm = [
		(self.br_rp, "left", 0),(self.br_rp, "right", 0),(self.br_rp,"top",10),
		(ok_btn, "bottom", 10),
		]
		
		attachControl =[
		(ok_btn, "top", 15, self.br_rp),
		]
		
		attachPosition =[
		(ok_btn, "left", 0, 45),
		(ok_btn, "right", 0, 50),
		]
		
		
		cmds.formLayout(form, edit=True, attachForm=attachForm, attachControl=attachControl, attachPosition=attachPosition)
		cmds.showWindow(self.cnjl_wnd)
	
	def _bakeRig_callback(self,a):
		
		prefix = cmds.textFieldGrp(self.br_rp, q=True, text=True )
		self.previewRig()
		
		ls = cmds.ls("ARCBUILDER__PRERIG__*")
		
		
		l = len("ARCBUILDER__PRERIG__")
		for nn in ls:
			newname = "%s%s"%(prefix, nn[l:])
			try:
				# This is because of mayas auto renaming of shapes and such
				cmds.rename(nn, newname)
			except:
				pass
		
		cmds.deleteUI(self.cnjl_wnd)
	
	
	def saveRigData(self, a):
		
		try:
			filename = cmds.fileDialog2(ff="ARC Rigdata (.arcrig) (*.arcrig)")[0]
		except:
			return
		
		if((filename == None) or (filename == "")):
			return
		jd_list = []
		for i in range(len(self.joints)):
			jd_list.append([self.joints[i][0], self.joints[i][1], cmds.xform(self.joints[i][2], q=True, ws=True, a=True, t=True)])
		fp = open(filename, "wb+")
		_pickle.dump([jd_list, self.bendy_iks, self.pv_iks, self.geo_aimc, self.geo_pc, self.geo_sspc], fp)
		fp.close()
	
	
	def loadRigData(self, a):
		
		try:
			filename = cmds.fileDialog2(fm=1, okc="Open", ff="ARC Rigdata (.arcrig) (*.arcrig)")[0]
		except:
			return
		
		if((filename == None) or (filename == "")):
			return
		fp = open(filename, "rb")
		datas =_pickle.load(fp)
		jd_list = datas[0]
		self.bendy_iks = datas[1]
		self.pv_iks = datas[2]
		self.geo_aimc = datas[3]
		self.geo_pc = datas[4]
		self.geo_sspc = datas[5]
		fp.close()
		
		
		for i in range(len(jd_list)):
			
			name = jd_list[i][0]
			parent = jd_list[i][1]
			nPos = jd_list[i][2]
			
			self.addLocator(name,parent,nPos)
		
		self.updateJointsList()
		self.updateIKsList()


import inspect
import os

arc_rig_utils_file = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
arc_rig_utils_start = inspect.currentframe().f_lineno

import maya.cmds as cmds

class arc_rig_utils:
	
	SELECTMODE_COM = 0
	SELECTMODE_OBJ = 1
	SELECTMODE_ROOT = 2
	SELECTMODE_LEAF = 3
	SELECTMODE_TEMP = 4
	SELECTMODE_HIER = 5
	SELECTMODE_PRES = 6
	
	
	def __init__(self):
		pass
	
	
	# set colour of object
	@staticmethod
	def setColor(nd, col):
		cmds.setAttr("%s.overrideEnabled"%nd, True)
		cmds.setAttr("%s.overrideColor"%nd, col)
	
	# returns currently selected mode
	@staticmethod
	def getSelectMode():
		if cmds.selectMode(q=True, co=True):
			return arc_rig_utils.SELECTMODE_COM
		if cmds.selectMode(q=True, o=True):
			return arc_rig_utils.SELECTMODE_OBJ
		if cmds.selectMode(q=True, r=True):
			return arc_rig_utils.SELECTMODE_ROOT
		if cmds.selectMode(q=True, l=True):
			return arc_rig_utils.SELECTMODE_LEAF
		if cmds.selectMode(q=True, t=True):
			return arc_rig_utils.SELECTMODE_TEMP
		if cmds.selectMode(q=True, h=True):
			return arc_rig_utils.SELECTMODE_HIER
		if cmds.selectMode(q=True, p=True):
			return arc_rig_utils.SELECTMODE_PRES
		return -1
	
	
	# sets selectmode according to SELECTMODE_ variables
	@staticmethod
	def setSelectMode(mode):
		
		#print mode
		# Python switch
		a = {
			arc_rig_utils.SELECTMODE_COM : "cmds.selectMode(co=True)",
			arc_rig_utils.SELECTMODE_OBJ : "cmds.selectMode(o=True)",
			arc_rig_utils.SELECTMODE_ROOT : "cmds.selectMode(r=True)",
			arc_rig_utils.SELECTMODE_LEAF : "cmds.selectMode(l=True)",
			arc_rig_utils.SELECTMODE_TEMP : "cmds.selectMode(t=True)",
			arc_rig_utils.SELECTMODE_HIER : "cmds.selectMode(h=True)",
			arc_rig_utils.SELECTMODE_PRES : "cmds.selectMode(p=True)",
		}.get(mode,"0")
		eval(a)
		
	
	
	@staticmethod
	def getSelectedLocations():
		r = []
		vs = cmds.ls(fl=1,os=1)
		for v in vs:
			r.append(cmds.xform(v, q=True, ws=True, a=True, t=True))
		return r
	
	@staticmethod
	def averagePosFromSelected():
		pts = arc_rig_utils.getSelectedLocations()
		l = len(pts)
		if l < 1:
			return None
		r = [0,0,0]
		for v in pts:
			r[0] += v[0]
			r[1] += v[1]
			r[2] += v[2]
		return [r[0]/l, r[1]/l, r[2]/l]
	
	@staticmethod
	def joint(owner=None, prefix="", name="", suffix="", p=(0,0,0), joint_size=0.5):
		
		jnt = cmds.joint(owner, p=p, n="%s%s%s"%(prefix, name, suffix), rad=joint_size)
		arc_rig_utils.freezeDelete(jnt)
		# Set prefered angle
		cmds.joint(jnt, e=True, spa=True, ch=True)
		return jnt
		
	
	
	#a = arc_rig_utils.pv_ikhandle("r_ik_gen", "myrigL_shoulder", "myrigL_wrist")
	#a = arc_rig_utils.pv_ikhandle("l_ik_gen", "myrigR_shoulder", "myrigR_wrist")
	#a = arc_rig_utils.pv_ikhandle("p_ik_gen", "myrigspine_1", "myrigspine_2")
	#a = arc_rig_utils.pv_ikhandle("n_ik_gen", "myrigneck", "myrighead")
	#a = arc_rig_utils.spline_ikhandle("bendy_ik_gen", "myrigspine_1", "myrigspine_4")
	# Default controller type is Cube
	@staticmethod
	def pv_ikhandle(name, startJoint, endJoint, controller_size=1.0, locatorSize=0.5, locatorOffset=[0,-1,-3], owner=None, color=14):
		
		ik = cmds.ikHandle(n=name, ap=True, sol="ikRPsolver", sj=startJoint, ee=endJoint)
		arc_rig_utils.setColor(ik[0],5)
		
		
		sj_loc = cmds.xform(startJoint, q=True, ws=True, a=True, t=True)
		ej_loc = cmds.xform(endJoint, q=True, ws=True, a=True, t=True)
		
		lc_pos = [ 
		(sj_loc[0]+ej_loc[0])/2.0 + locatorOffset[0],
		(sj_loc[1]+ej_loc[1])/2.0 + locatorOffset[1],
		(sj_loc[2]+ej_loc[2])/2.0 + locatorOffset[2]
		]
		
		
		# Create Locator for pole vector
		poleVector_loc = cmds.spaceLocator(n="%s_pv_loc"%name, p=(0,0,0))[0]
		cmds.xform(poleVector_loc, s=[locatorSize,locatorSize,locatorSize])
		cmds.xform(poleVector_loc, ws=True, a=True, t=lc_pos)
		arc_rig_utils.setColor(poleVector_loc, 13)
		
		
		# Create pole vector / controller
		pv = cmds.poleVectorConstraint(poleVector_loc, ik[0], weight=1);
		ctrl = arc_rig_utils.makeCube(name="%s_ctrl"%name, size=controller_size, pivot=ej_loc)
		arc_rig_utils.setColor(ctrl, color)
		
		
		# Parent to controller
		cmds.parent(poleVector_loc, ctrl)
		cmds.parent(ik[0], ctrl)
		
		
		#Lock /Hide IK transforms
		lockList = ["%s.%s"%(ik[0],v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in lockList:
			arc_rig_utils.lockAttr(v)
			arc_rig_utils.hideAttr(v)
		
		_name = name
		if _name.startswith("ARCBUILDER__PRERIG__"):
			_name = _name[len("ARCBUILDER__PRERIG__"):]
		
		# Add IK/FK switch, also hide the expression
		arc_rig_utils.addEnumAttr(ctrl, "IkFkMode", ["IK","FK"])
		expr = arc_rig_utils.createExpression(name="%s_expr"%name, owner=ik[0], expr="ikBlend = 1-float(%s.IkFkMode)"%ctrl)
		
		"""
		# Hide Controller in FK mode
		expr = arc_rig_utils.createExpression(name="%s_vis"%ctrl, owner=ctrl, expr="visibility = 1.0-float(IkFkMode)")
		try:
			arc_rig_utils.hide_AE(expr)
		except:
			pass
		"""
		
		if owner != None:
			
			arc_rig_utils.addFloatAttr(owner, "%s_IkFkMode"%_name, 1.0)
			arc_rig_utils.connectAttr("%s.IkFkMode"%ctrl,"%s.%s_IkFkMode"%(owner,_name))
		
		# Set Joint colour
		arc_rig_utils.setColor(startJoint,20)
		
		
		arc_rig_utils.freezeDelete(ik[0])
		arc_rig_utils.freezeDelete(ctrl)
		arc_rig_utils.freezeDelete(pv)
		arc_rig_utils.freezeDelete(poleVector_loc)
		return [ik,ctrl,pv,poleVector_loc]
	
	@staticmethod
	def spline_ikhandle(name, startJoint, endJoint, numberOfJoints=1, rootTwist=True, createControllers=True, controllerSize=1.0, owner=None):
		try:
			ik=cmds.ikHandle(n=name , sol="ikSplineSolver", pcv=False, ns=numberOfJoints, sj=startJoint, ee=endJoint, rtm=rootTwist)
		except:
			print "Error: Could not create IK!"
			return
		
		ik[2]=cmds.rename(ik[2], "%s_curve"%name)
		clusters=[]
		
		grp = cmds.group(n="%s_grp"%name, empty=True)
		
		# Add IK/FK Switch controls
		arc_rig_utils.addEnumAttr(grp, "IkFkMode", ["IK","FK"])
		arc_rig_utils.hideAttr("%s.IkFkMode"%grp)
		expr = arc_rig_utils.createExpression(name="%s_expr"%name, owner=ik[0], expr="ikBlend = 1.0-float(%s.IkFkMode)"%grp)
		
		_name = name
		if _name.startswith("ARCBUILDER__PRERIG__"):
			_name = _name[len("ARCBUILDER__PRERIG__"):]
		
		if owner != None:
			
			arc_rig_utils.addEnumAttr(owner, "%s_IkFkMode"%_name, ["IK","FK"])
			arc_rig_utils.connectAttr("%s.%s_IkFkMode"%(owner,_name), "%s.IkFkMode"%grp)
		
		try:
			arc_rig_utils.hide_AE(expr)
		except:
			pass
			
		cmds.parent(ik[0], grp)
		cmds.parent(ik[2], grp)
		
		arc_rig_utils.freezeDelete(ik[0])
		arc_rig_utils.freezeDelete(ik[2])
		
		for i in xrange(0, numberOfJoints+3):
			cluster = cmds.cluster("%s.cv[%i]"%(ik[2],i), rel=True, n="%s_cluster%i"%(name,i), envelope=1)
			
			j = i
			if j == 0:
				j += 1
			
			pos = cmds.xform("%s.cv[%i]"%(ik[2],i), q=True, ws=True, a=True, t=True)
			
			# Get controller direction
			cur_cv = cmds.xform("%s.cv[%i]"%(ik[2],j), q=True, ws=True, a=True, t=True)
			pre_cv = cmds.xform("%s.cv[%i]"%(ik[2],j-1), q=True, ws=True, a=True, t=True)
			
			controller_normal_x = cur_cv[0]-pre_cv[0]
			controller_normal_y = cur_cv[1]-pre_cv[1]
			controller_normal_z = cur_cv[2]-pre_cv[2]
			controller_n_max = max(max(abs(controller_normal_x), abs(controller_normal_y)),abs(controller_normal_z))
			
			
			controller_normal = [controller_normal_x/controller_n_max, controller_normal_y/controller_n_max, controller_normal_z/controller_n_max]
			root_ctrl = False
			
			
			controller = cmds.circle(n="%s_%i_ctrl" %(name, i), r=controllerSize, ch=False, normal=controller_normal, center=pos, fc=True)[0]
			cmds.xform(controller, cp=True)
			
			"""
			# Hide when FK mode is on
			expr = arc_rig_utils.createExpression(name="%s_vis"%controller, owner=controller, expr="visibility = 1.0-float(%s.IkFkMode)"%grp)
			try:
				arc_rig_utils.hide_AE(expr)
			except:
				pass
			"""
			
			# Hide Cluster
			cmds.hide(cluster)
			
			if root_ctrl:
				cmds.hide(controller)
			
			# Group Cluster
			cmds.parent(cluster,controller)
			
			# When its relative we need to OVER drive it
			arc_rig_utils.connectAttr("%s.translate"%controller,"%s.translate"%cluster[1])
			
			# Group controller
			cmds.parent(controller, grp)
			
			
			# Set Controller colour
			arc_rig_utils.setColor(controller,8)
			
			arc_rig_utils.freezeDelete(cluster)
			arc_rig_utils.freezeDelete(controller)
			
			clusters.append([cluster,controller])
		
		ik.append(clusters)
		
		lockList = ["%s.%s"%(grp,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in lockList:
			arc_rig_utils.lockAttr(v)
			arc_rig_utils.hideAttr(v)
		
		# Set Joint colour
		arc_rig_utils.setColor(startJoint,23)
		
		return [ik,grp]
	
	
	@staticmethod
	def createAimConstraint(name, parent, controller_size=0.1, controller_offset=[0,0,2], parent_joint=None, aimsback=True, color=4):
		parent_pos = cmds.xform(parent,q=True, ws=True, a=True, t=True)
		joint_pos = cmds.xform(parent_joint,q=True, ws=True, a=True, t=True)
		
		#ctrl_pos = [parent_pos[i]+controller_offset[i] for i in range(3)]
		ctrl_pos = [(parent_pos[i]-joint_pos[i])+controller_offset[i] for i in range(3)]
		controller = cmds.circle(n="%s_%s_aim_ctrl" %(name,parent), r=controller_size, ch=False, normal=[1,0,0], center=[0,0,0], fc=True)[0]
		cmds.xform(controller, t=ctrl_pos)
		
		
		# Unlock parent
		unlockList = ["%s.%s"%(parent,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in unlockList:
			arc_rig_utils.unlockAttr(v)
		
		ac = cmds.aimConstraint(controller,parent, mo=True, weight=1, aimVector=[0,0,1], upVector=[0,1,0] ,worldUpType="vector", worldUpVector=[0,1,0])[0]
		# Doing this manually
		for v in ("offsetX","offsetY","offsetZ"):
			cmds.setAttr("%s.%s"%(ac,v), 0)
		
		arc_rig_utils.setColor(controller, color)
		
		if(aimsback):
			arc_rig_utils.connectAttr("%s.constraintRotate"%ac,"%s.rotate"%controller)
		
		# Make line from eye to controller
		
		ln = arc_rig_utils.makeSolidShape([parent_pos, ctrl_pos])
		ln = cmds.rename(ln,"%s_%s_guide_ctrl"%(name,parent))
		arc_rig_utils.connectAttr("%s.translate"%parent, "%s.cv[0]"%ln)
		arc_rig_utils.connectAttr("%s.translate"%controller, "%s.cv[1]"%ln)
		
		arc_rig_utils.setColor(ln,color)
		
		# Add checbox so you can turn the guide off and on
		arc_rig_utils.addBooleanAttr(controller, "ShowGuide", True)
		expr = arc_rig_utils.createExpression(name="%s_expr"%ln, owner=ln, expr="visibility = (%s.ShowGuide)"%controller)
		
		# Create Group
		grp = cmds.group(n=name, empty=True)
		cmds.parent(controller, grp)
		cmds.parent(ln, grp)
		
		
		# Dont let people move the guide around themselves
		lockList = ["%s.%s"%(ln,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in lockList:
			arc_rig_utils.lockAttr(v)
			arc_rig_utils.hideAttr(v)
		
		# Dont let people move the group around themselves
		lockList = ["%s.%s"%(grp,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in lockList:
			arc_rig_utils.lockAttr(v)
			arc_rig_utils.hideAttr(v)
		
		
		# Parenting the controller and a constraint to the joint
		if parent_joint != None:
			# This is so the object follows the joint
			cmds.parent(grp,  parent_joint)
			pc = cmds.parentConstraint(parent_joint, parent, mo=True, skipRotate=["x","y","z"], weight=1)[0]
			arc_rig_utils.connectAttr("%s.target[0].targetOffsetTranslate"%pc, "%s.cv[0]"%ln)
			
		return [controller, ln, grp]
			
			
	@staticmethod
	def createPointController(name, points, controller_size=0.05, distance=0.2, parent_joint=None, color=6):
		normals = cmds.polyNormalPerVertex(points, q=True, normalXYZ=True)
		
		n_normals=[0,0,0]
		for i in range(len(normals)/3):
			n_normals[0] += normals[i*3]
			n_normals[1] += normals[i*3+1]
			n_normals[2] += normals[i*3+2]
		
		n_normals[0] /= len(normals)/3;
		n_normals[1] /= len(normals)/3;
		n_normals[2] /= len(normals)/3;
		
		parent_pos = cmds.xform(points,q=True, ws=True, a=True, t=True)
		ctrl_pos = [parent_pos[i]+(n_normals[i]*distance) for i in range(3)]
		
		#controller = cmds.circle(n="%s_pc_ctrl" %(name), r=controller_size, ch=False, normal=n_normals, center=[0,0,0], fc=True)[0]
		
		# Creating a nurbs sphere, then removing it from the shading list
		# Then super making sure it wont get rendered
		# Followed by colouring it
		controller = cmds.sphere(n="%s_pc_ctrl" %(name), r=controller_size, po=0, ax=[0,1,0] , nsp=3, ch=0)[0]
		
		controller_sh = cmds.listRelatives(controller, children=True)[0]
		shadingGroupPlug = cmds.listConnections("%s.instObjGroups[0]"%controller_sh, p=True)[0]
		arc_rig_utils.disconnectAttr(shadingGroupPlug)
		arc_rig_utils.setColor(controller_sh, color)
		for v in ["castsShadows", "receiveShadows", "motionBlur", "primaryVisibility", "visibleInReflections", "visibleInRefractions"]:
			cmds.setAttr("%s.%s"%(controller_sh,v), 0)
		
		
		cmds.xform(controller, t=ctrl_pos)
		cmds.xform(controller, cp=True)
		
		arc_rig_utils.freezeDelete(controller)
		
		
		
		clus = cmds.cluster(points, rel=False, n="%s_cluster"%(name), envelope=1)
		
		# We are being very VERY extra here, instead of making a parent constraint,
		# in the hopes of keeping a clean outliner, we're using vectorProduct to
		# drive the cluster, which drives the point
		# ...I know
		vecProd = cmds.createNode("vectorProduct", n="%s_vecprod"%name)
		cmds.setAttr("%s.operation"%vecProd, 3);
		arc_rig_utils.connectAttr("%s.worldMatrix[0]"%controller, "%s.matrix"%vecProd)
		arc_rig_utils.connectAttr("%s.translate"%controller, "%s.input1"%vecProd)
		arc_rig_utils.connectAttr("%s.output"%vecProd, "%s.translate"%clus[1])
		
		cmds.setAttr("%s.inheritsTransform"%clus[1], 0);
		
		
		#arc_rig_utils.setColor(controller, color)
		
		if parent_joint != None:
			cmds.parent(controller,  parent_joint)
			arc_rig_utils.freezeDelete(controller)
		
		cmds.hide(clus[1])
		
		return [controller,clus]
	
	
	@staticmethod
	def createSoftPointController(name, points, controller_size=0.2, distance=0.2, fallout_radius=1, owner=None, color=9):
		normals = cmds.polyNormalPerVertex(points, q=True, normalXYZ=True)
		
		point_owner = points[:points.index(".")]
		
		n_normals=[0,0,0]
		for i in range(len(normals)/3):
			n_normals[0] += normals[i*3]
			n_normals[1] += normals[i*3+1]
			n_normals[2] += normals[i*3+2]
		
		n_normals[0] /= len(normals)/3;
		n_normals[1] /= len(normals)/3;
		n_normals[2] /= len(normals)/3;
		
		parent_pos = cmds.xform(points,q=True, ws=True, a=True, t=True)
		owner_pos = cmds.xform(owner,q=True, ws=True, a=True, t=True)
		ctrl_pos = [parent_pos[i]+(n_normals[i]*distance) - owner_pos[i] for i in range(3)]
		controller = cmds.circle(n="%s_softpc_ctrl"%(name), r=controller_size, ch=False, normal=n_normals, center=[0,0,0], fc=True)[0]
		cmds.xform(controller, t=ctrl_pos)
		
		
		sm = cmds.softMod(point_owner, n="%s_sm"%name, fc=parent_pos, fr=fallout_radius, rel=True)
		
		cmds.hide(sm[0])
		cmds.hide(sm[1])
		
		
		arc_rig_utils.freezeDelete(controller)
		arc_rig_utils.connectAttr("%s.translate"%controller, "%s.translate"%sm[1])
		arc_rig_utils.connectAttr("%s.rotate"%controller, "%s.rotate"%sm[1])
		arc_rig_utils.connectAttr("%s.scale"%controller, "%s.scale"%sm[1])
		
		arc_rig_utils.setColor(controller, color)
		
		grp = cmds.group(n="%s_aimsback_grp"%name, empty=True)
		
		# Group will move around with the point owner geo
		arc_rig_utils.connectAttr("%s.rotatePivot"%point_owner, "%s.rotatePivot"%grp)
		arc_rig_utils.connectAttr("%s.rotatePivotTranslate"%point_owner, "%s.rotatePivotTranslate"%grp)
		arc_rig_utils.connectAttr("%s.scalePivot"%point_owner, "%s.scalePivot"%grp)
		arc_rig_utils.connectAttr("%s.scalePivotTranslate"%point_owner, "%s.scalePivotTranslate"%grp)
		arc_rig_utils.connectAttr("%s.rotate"%point_owner, "%s.rotate"%grp)
		arc_rig_utils.connectAttr("%s.translate"%point_owner, "%s.translate"%grp)
		arc_rig_utils.connectAttr("%s.scale"%point_owner, "%s.scale"%grp)
		
		
		grp2 = cmds.group(n="%s_grp"%name, empty=True)
		
		
		# Dont let people move the group around themselves
		lockList = ["%s.%s"%(grp,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in lockList:
			arc_rig_utils.lockAttr(v)
			arc_rig_utils.hideAttr(v)
			
		# Dont let people move the group around themselves
		lockList = ["%s.%s"%(grp2,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in lockList:
			arc_rig_utils.lockAttr(v)
			arc_rig_utils.hideAttr(v)
		
		
		# Add Controller
		arc_rig_utils.addFloatAttr(controller, "Falloff", fallout_radius)
		arc_rig_utils.connectAttr("%s.Falloff"%controller, "%s.falloffRadius"%sm[0])
		
		
		cmds.parent(sm[1], grp2)
		cmds.parent(controller,grp)
		cmds.parent(grp,grp2)
		
		arc_rig_utils.freezeDelete(controller)
		
		
		return [controller,grp,grp2]
		

	# Linear interpolation (number1, number2, 0.0-1.0 percentage)
	@staticmethod
	def lerp(v1, v2, p):
		return v1 + (v2-v1)*p
	
	
	@staticmethod
	def makeSolidShape(pnts=[]):
		k = [i for i in range(len(pnts))]
		return cmds.curve(d=1, ws=True, p=pnts, k=k)
		
	
	@staticmethod
	def makeCubeShape(size=1, pivot=[0,0,0]):
		ret = []
		s = float(size)*.5
		x=float(pivot[0])
		y=float(pivot[1])
		z=float(pivot[2])
		# In all honesty, I brute forced the second line of this 
		for v in (-s,s):
			for a in [(s+x,v+y,-s+z), (-s+x,v+y,-s+z), (-s+x,v+y,s+z), (s+x,v+y,s+z), (s+x,v+y,-s+z),
				(-v+x,s+y,-s+z), (-v+x,-s+y,-s+z), (-v+x,-s+y,s+z), (-v+x,s+y,s+z), (-v+x,s+y,-s+z)]:
				ret.append(a)
		
		return ret
	
	@staticmethod
	def makeCube(name=None, size=1, pivot=[0,0,0], scale_top=1.0, scale_bottom=1.0, scale_left=1.0, scale_right=1.0, scale_front=1.0, scale_back=1.0):
		r= arc_rig_utils.makeSolidShape(arc_rig_utils.makeCubeShape(size,pivot))
		
		cmds.xform(r,cp=True)
		cmds.xform(["%s.cv[5]"%r, "%s.cv[8:15]"%r, "%s.cv[18:19]"%r],s=[scale_top,1,scale_top])
		cmds.xform(["%s.cv[0:4]"%r, "%s.cv[6:7]"%r, "%s.cv[16:17]"%r],s=[scale_bottom,1,scale_bottom])
		
		cmds.xform(["%s.cv[0]"%r, "%s.cv[3:10]"%r, "%s.cv[13:14]"%r],s=[1,scale_left,scale_left])
		cmds.xform(["%s.cv[1:2]"%r, "%s.cv[11:12]"%r, "%s.cv[15:19]"%r],s=[1,scale_right,scale_right])
		
		cmds.xform(["%s.cv[2:3]"%r, "%s.cv[7:8]"%r, "%s.cv[12:13]"%r, "%s.cv[17:18]"%r],s=[scale_front,scale_front,1])
		cmds.xform(["%s.cv[0:1]"%r, "%s.cv[4:6]"%r, "%s.cv[9:11]"%r, "%s.cv[14:16]"%r, "%s.cv[19]"%r],s=[scale_back,scale_back,1])
		
		cmds.xform(r,cp=True)
		
		# Renaming the shapes, becasue looks pretty...
		shapes = cmds.listRelatives(r)
		cmds.rename(shapes[0], "%s_nc"%r)
		
		if(name != None):
			r = cmds.rename(r, name)
		
		return r
	
	# [-1,0,0],[0,0,-1],[1,0,0],[0,0,1],[-1,0,0]
	@staticmethod
	def makeDiamondShape(size=1, pivot=[0,0,0]):
		ret=[]
		x=pivot[0]
		y=pivot[1]
		z=pivot[2]
		s = size*.5
		for v in (-s,s):
			ret.append([x+v, y, z])
			ret.append([x, y, z+v])
		ret.append([-s+x, y, z])
		return ret
	
	@staticmethod
	def makeDiamond(name=None, size=1, pivot=[0,0,0]):
		r = arc_rig_utils.makeSolidShape(arc_rig_utils.makeDiamondShape(size,pivot))
		shapes = cmds.listRelatives(r)
		cmds.rename(shapes[0], "%s_nc"%r)
		
		if name != None:
			r= cmds.rename(r,name)
			
		return r
		
	@staticmethod
	def makeMultipleDiamond(name=None, sizes=[0.5,1], pivot=[0,0,0]):
		s = []
		for v in sizes:
			s = s + arc_rig_utils.makeDiamondShape(v,pivot)
		r = arc_rig_utils.makeSolidShape(s)
		shapes = cmds.listRelatives(r)
		cmds.rename(shapes[0], "%s_nc"%r)
		
		if name != None:
			r= cmds.rename(r,name)
			
		return r
	
	# 5x size
	#[-1,0,-1],[-3,0,-1],[-3,0,-2],[-5,0,0],[-3,0,2],[-3,0,1],[-1,0,1],[-1,0,3],[-2,0,3],[0,0,5],[2,0,3],[1,0,3],[1,0,1],[3,0,1],[3,0,2],[5,0,0],[3,0,-2],[3,0,-1],[1,0,-1],[1,0,-3],[2,0,-3],[0,0,-5],[-2,0,-3],[-1,0,-3],[-1,0,-1]
	# type 0 = edges of arrows seperated
	# type 1 = edges of arrows connected
	# type 2 = edges of lines pointing out
	# type 3 = edges of lines become square
	@staticmethod
	def makeCompassShape(size=1, pivot=[0,0,0], type=0):
		base = [[-0.2,0,-0.2],[-0.6,0,-0.2],[-0.6,0,-0.4],[-1.0,0,0],[-0.6,0,0.4],[-0.6,0,0.2],[-0.2,0,0.2],[-0.2,0,0.6],[-0.4,0,0.6],[0,0,1.0],[0.4,0,0.6],[0.2,0,0.6],[0.2,0,0.2],[0.6,0,0.2],[0.6,0,0.4],[1.0,0,0],[0.6,0,-0.4],[0.6,0,-0.2],[0.2,0,-0.2],[0.2,0,-0.6],[0.4,0,-0.6],[0,0,-1.0],[-0.4,0,-0.6],[-0.2,0,-0.6],[-0.2,0,-0.2]]
		
		if type == 1:
			base[0][0] -= 0.2
			base[0][2] -= 0.2
			base[24][0] -= 0.2
			base[24][2] -= 0.2
			base[18][0] += 0.2
			base[18][2] -= 0.2
			base[12][0] += 0.2
			base[12][2] += 0.2
			base[6][0] -= 0.2
			base[6][2] += 0.2
		
		if type == 2:
			base[0][0] -= 0.3
			base[0][2] -= 0.3
			base[24][0] -= 0.3
			base[24][2] -= 0.3
			base[18][0] += 0.3
			base[18][2] -= 0.3
			base[12][0] += 0.3
			base[12][2] += 0.3
			base[6][0] -= 0.3
			base[6][2] += 0.3
			
		if type == 3:
			base[0][0] -= 0.4
			base[0][2] -= 0.4
			base[24][0] -= 0.4
			base[24][2] -= 0.4
			base[18][0] += 0.4
			base[18][2] -= 0.4
			base[12][0] += 0.4
			base[12][2] += 0.4
			base[6][0] -= 0.4
			base[6][2] += 0.4
			
			
		
		for v in base:
			v[0] = (v[0]*size)+pivot[0]
			v[1] = (v[1]*size)+pivot[1]
			v[2] = (v[2]*size)+pivot[2]
		return base
	
	@staticmethod
	def makeCompass(name=None, size=1, pivot=[0,0,0], type=0):
		r = arc_rig_utils.makeSolidShape(arc_rig_utils.makeCompassShape(size,pivot,type))
		shapes = cmds.listRelatives(r)
		cmds.rename(shapes[0], "%s_nc"%r)
		
		if name != None:
			r= cmds.rename(r,name)
		return r
	
	@staticmethod
	def makeGlobalControllerObject(name=None, size=1.0, pivot=[0,0,0], cols=[23,19]):
		compass_outer = arc_rig_utils.makeCompass(name, size, pivot, type=2)
		compass_inner = cmds.circle(n="%s_inner"%(compass_outer), r=size*.5, ch=False, normal=[0,1,0], center=[0,0,0], fc=True)[0]
		
		lockList = ["%s.%s"%(compass_inner,v) for v in ("tx","ty","tz","rx","ry","rz","sx","sy","sz")]
		for v in lockList:
			arc_rig_utils.lockAttr(v)
			arc_rig_utils.hideAttr(v)
		
		cmds.parent(compass_inner, compass_outer)
		arc_rig_utils.setColor(compass_outer, cols[0])
		arc_rig_utils.setColor(compass_inner, cols[1])
		arc_rig_utils.freezeDelete(compass_outer)
		arc_rig_utils.freezeDelete(compass_inner)
		return [compass_outer, compass_inner]
	
	
	@staticmethod 
	def createExpression(owner, expr, name=None):
		return cmds.expression(n=name, s=expr, o=owner, ae=True, uc="all")
	
	@staticmethod
	def hide_AE(name):
		mel.eval("hideNodeTypeAEFilter \"%s\";"%name)
	
	@staticmethod
	def hideAttr(name):
		cmds.setAttr(name, keyable=False, channelBox=False)
	
	@staticmethod
	def lockAttr(name):
		cmds.setAttr(name, lock=True)
	
	@staticmethod
	def unlockAttr(name):
		cmds.setAttr(name, lock=False)
		
	@staticmethod
	def connectAttr(_from, _to):
		cmds.connectAttr(_from, _to, f=True)
		
	@staticmethod
	def disconnectAttr(plug):
		try:
			dest = cmds.connectionInfo(plug, sourceFromDestination=True)
			cmds.disconnectAttr(dest, plug)
		except:
			pass
		
	@staticmethod
	def addEnumAttr(parent, name, vals):
		cmds.addAttr(parent, ln=name, at="enum", en=":".join(vals)+":");
		cmds.setAttr("%s.%s"%(parent,name), e=True, keyable=True)
	
	@staticmethod
	def addBooleanAttr(parent, name, val):
		cmds.addAttr(parent, ln=name, at="bool")
		cmds.setAttr("%s.%s"%(parent,name), val, e=True, keyable=True)
	
	@staticmethod
	def addFloatAttr(parent, name, val):
		cmds.addAttr(parent, ln=name, at="double")
		cmds.setAttr("%s.%s"%(parent,name), val, e=True, keyable=True)
	
	@staticmethod
	def freezeTransforms(obj, s=1, t=1, r=1, jo=1, n=0):
		try:
			cmds.makeIdentity(obj, apply=True, t=t, r=r, s=s, jo=jo, n=n)
		except Exception as e:
			print e
	
	@staticmethod
	def deleteHistroy(obj):
		cmds.DeleteHistory(obj)
		cmds.delete(obj,ch=True)
		
	@staticmethod
	def freezeDelete(obj, s=1, t=1, r=1, jo=1, n=0):
		arc_rig_utils.deleteHistroy(obj)
		arc_rig_utils.freezeTransforms(obj,s,t,r,jo,n)
	
	@staticmethod
	def getUnfrozenWSPos(obj):
		cur = cmds.xform(obj, q=True, ws=True, a=True, rp=True)
		cmds.xform(obj, cp=True)
		new = cmds.xform(obj, q=True, ws=True, a=True, rp=True)
		cmds.xform(obj, ws=True, rp=cur)
		return [cur[0]+new[0], cur[1]+new[1], cur[2]+new[2]]
